from dataclasses import dataclass, field
from datetime import date
from enum import Enum
from pathlib import Path
from typing import Annotated, Any, Literal

import pytest

from serieux import Serieux
from serieux.auto import Auto, Call
from serieux.ctx import Context
from serieux.exc import ValidationError
from serieux.features.clargs import CommandLineArguments, parse_cli
from serieux.features.fromfile import IncludeFile, WorkingDirectory
from serieux.features.interpol import Environment
from serieux.features.tagset import TagDict, Tagged, TaggedUnion

from ..definitions import Defaults, Job, Point, Worker

deserialize = (Serieux + IncludeFile)().deserialize

datapath = Path(__file__).parent.parent / "data"


@dataclass
class Person:
    # Name of the person
    name: str

    # Age of the person
    age: int


@dataclass
class HeroProfile:
    beautiful: bool = True
    has_superpowers: bool = False
    loves_adventure: bool = True


class Material(str, Enum):
    WOOD = "wood"
    BRICK = "brick"
    CONCRETE = "concrete"
    STEEL = "steel"
    GLASS = "glass"


@dataclass
class House:
    # Date at which the house was built
    built: date

    # Building material
    material: Material


def test_simple():
    result = deserialize(
        Person,
        CommandLineArguments(["--name", "Jon", "--age", "27"]),
        Context(),
    )
    assert result == Person(name="Jon", age=27)


def test_variables():
    result = deserialize(
        Person,
        CommandLineArguments(["--name", "Jon", "--age", "${env:AGE}"]),
        Environment(environ={"AGE": "33"}),
    )
    assert result == Person(name="Jon", age=33)


def test_help(capsys, file_regression):
    with pytest.raises(SystemExit):
        deserialize(Person, CommandLineArguments(["-h"]), Context())
    captured = capsys.readouterr()
    file_regression.check(captured.out + "\n=====\n" + captured.err)


def test_booleans():
    def f(*argv):
        return deserialize(HeroProfile, CommandLineArguments(argv), Context())

    assert f() == HeroProfile(True, False, True)
    assert f("--no-beautiful") == HeroProfile(False, False, True)
    assert f("--no-beautiful", "--has-superpowers", "--no-loves-adventure") == HeroProfile(
        False, True, False
    )


def test_lists():
    def f(*argv):
        return deserialize(Defaults, CommandLineArguments(argv), Context())

    assert f("--name", "Tyrone") == Defaults(name="Tyrone", aliases=[], cool=False)
    assert f("--name", "Tyrone", "--aliases", "Ty", "Ro", "Ne") == Defaults(
        name="Tyrone", aliases=["Ty", "Ro", "Ne"], cool=False
    )


@dataclass
class Defaults2:
    name: str
    aliases: list[str] = field(default_factory=list, metadata={"option": "-a", "action": "append"})
    cool: bool = field(default=False, kw_only=True)


def test_lists_append():
    def f(*argv):
        return deserialize(Defaults2, CommandLineArguments(argv), Context())

    assert f("--name", "Tyrone") == Defaults2(name="Tyrone", aliases=[], cool=False)
    assert f("--name", "Tyrone", "-a", "Ty", "-a", "Ro", "-a", "Ne") == Defaults2(
        name="Tyrone", aliases=["Ty", "Ro", "Ne"], cool=False
    )


def test_misc_types():
    def f(*argv):
        return deserialize(House, CommandLineArguments(argv), Context())

    assert f("--built", "2025-01-01", "--material", "brick") == House(
        built=date(2025, 1, 1), material=Material.BRICK
    )
    with pytest.raises(SystemExit):
        f("--built", "2025-01-01", "--material", "invalid")


@dataclass
class Eat:
    """Stuffing your mouth."""

    food: str = None

    def do(self):
        return f"I eat {self.food}"


@dataclass
class Sleep:
    """Stuffing your brain with dreams."""

    hours: int

    def do(self):
        return f"I sleep {self.hours} hours"


@dataclass
class Act:
    """Do stuff!"""

    # What to do
    command: TaggedUnion[Eat, Sleep]  # noqa: F821

    # Do we do it fast?
    fast: bool = field(default=False, metadata={"alias": "-f"})

    def text(self):
        return self.command.do() + (" fast" if self.fast else "")


@dataclass
class TalkingPoint(Point):
    def text(self):
        return f"Hi I am at ({self.x}, {self.y})"


def test_subcommands():
    def do(*args):
        result = deserialize(
            TaggedUnion[Eat, Sleep],
            CommandLineArguments(args),
            Context(),
        )
        return result.do()

    assert do("eat", "--food", "jam") == "I eat jam"
    assert do("sleep", "--hours", "8") == "I sleep 8 hours"


def test_single_subcommand():
    def text(*args):
        result = deserialize(
            Tagged[Eat],
            CommandLineArguments(args),
            Context(),
        )
        return result.do()

    assert text("eat", "--food", "jam") == "I eat jam"


def test_args_plus_subcommands():
    def text(*args):
        result = deserialize(Act, CommandLineArguments(args), Context())
        return result.text()

    assert text("--fast", "eat", "--food", "jam") == "I eat jam fast"
    assert text("-f", "eat", "--food", "jam") == "I eat jam fast"
    assert text("sleep", "--hours", "8") == "I sleep 8 hours"


def test_subcommands_help(capsys, file_regression):
    with pytest.raises(SystemExit):
        deserialize(Act, CommandLineArguments(["-h"]), Context())
    captured = capsys.readouterr()
    file_regression.check(captured.out + "\n=====\n" + captured.err)


def test_subcommand_help(capsys, file_regression):
    with pytest.raises(SystemExit):
        deserialize(Act, CommandLineArguments(["eat", "-h"]), Context())
    captured = capsys.readouterr()
    file_regression.check(captured.out + "\n=====\n" + captured.err)


def test_sub_subcommands():
    def text(*args):
        result = deserialize(
            Tagged[Act, "act"] | Tagged[TalkingPoint, "point"],
            CommandLineArguments(args),
            Context(),
        )
        return result.text()

    assert text("act", "eat", "--food", "jam") == "I eat jam"
    assert text("act", "--fast", "eat", "--food", "jam") == "I eat jam fast"
    assert text("act", "sleep", "--hours", "8") == "I sleep 8 hours"
    assert text("point", "-x", "1", "-y", "2") == "Hi I am at (1, 2)"


@dataclass
class Eater:
    command: TaggedUnion[Eat]

    def text(self):
        return self.command.do() + " only"


def test_one_sub_subcommands():
    def text(*args):
        result = deserialize(
            TaggedUnion[Eater],
            CommandLineArguments(args),
            Context(),
        )
        return result.text()

    assert text("eater", "eat", "--food", "jam") == "I eat jam only"


@dataclass
class Word:
    word: str = field(metadata={"positional": True})


def test_positional():
    result = deserialize(Word, CommandLineArguments(["amazing"]), Context())
    assert result == Word(word="amazing")


@dataclass
class Wordz:
    # [positional: ...]
    wordz: list[str]


def test_positional_multiple():
    result = deserialize(Wordz, CommandLineArguments(["hello", "everyone", "--cool"]), Context())
    assert result == Wordz(wordz=["hello", "everyone", "--cool"])


@dataclass
class Sentence:
    # [nargs: *]
    sentence: str


def test_nargs_on_str():
    result = deserialize(Sentence, CommandLineArguments(["--sentence", "Hi", "guys!"]), Context())
    assert result == Sentence(sentence="Hi guys!")


@dataclass
class Duck:
    # [option: -q]
    quacks: int


def test_replace_option():
    result = deserialize(Duck, CommandLineArguments(["-q", "7"]), Context())
    assert result == Duck(quacks=7)


def test_mapping():
    cla = CommandLineArguments(
        ["--tit", "Inspector", "--yar", "35000", "-n", "Gunther"],
        mapping={"job.title": "--tit", "job.yearly_pay": "--yar", "name": "-n"},
    )
    result = deserialize(Worker, cla, Context())
    assert result == Worker(name="Gunther", job=Job(title="Inspector", yearly_pay=35000))


def test_recursive():
    cla = CommandLineArguments(
        ["--name", "Gunther", "--title", "Inspector", "--yearly-pay", "35000"]
    )
    result = deserialize(Worker, cla, WorkingDirectory(datapath))
    assert result == Worker(name="Gunther", job=Job(title="Inspector", yearly_pay=35000))


def test_mapping_config_file():
    def deserialize_cli(args):
        cla = CommandLineArguments(
            args,
            mapping={"": {"auto": True, "option": "--config"}},
        )
        return deserialize(Worker, cla, WorkingDirectory(datapath))

    result = deserialize_cli(["--config", "worker.yaml"])
    assert result == Worker(name="Hagrid", job=Job(title="Vagrant", yearly_pay=10))

    result = deserialize_cli(["--config", "worker.yaml", "--title", "Inspector"])
    assert result == Worker(name="Hagrid", job=Job(title="Inspector", yearly_pay=10))

    with pytest.raises(ValidationError, match="there was no such file"):
        deserialize_cli(["--config", "whatever.yaml", "--title", "Inspector"])

    result = deserialize_cli(
        ["--config", '{"name":"Hagrid","job":{"title":"Vagrant","yearly_pay":10}}']
    )
    assert result == Worker(name="Hagrid", job=Job(title="Vagrant", yearly_pay=10))


def test_mapping_plus_config_file():
    def deserialize_cli(args):
        cla = CommandLineArguments(
            args,
            mapping={
                "job.title": {"option": "--title", "required": False},
                "": {"option": "--config"},
            },
        )
        return deserialize(Worker, cla, WorkingDirectory(datapath))

    result = deserialize_cli(["--config", "worker.yaml"])
    assert result == Worker(name="Hagrid", job=Job(title="Vagrant", yearly_pay=10))

    result = deserialize_cli(["--config", "worker.yaml", "--title", "Inspector"])
    assert result == Worker(name="Hagrid", job=Job(title="Inspector", yearly_pay=10))

    with pytest.raises(ValidationError, match="there was no such file"):
        deserialize_cli(["--config", "whatever.yaml", "--title", "Inspector"])

    result = deserialize_cli(
        ["--config", '{"name":"Hagrid","job":{"title":"Vagrant","yearly_pay":10}}']
    )
    assert result == Worker(name="Hagrid", job=Job(title="Vagrant", yearly_pay=10))


def test_parse_cli():
    result = parse_cli(
        root_type=Worker,
        mapping={
            "name": "-n",
            "job.title": "--tit",
            "job.yearly_pay": "--yar",
        },
        argv=["-n", "Gunther", "--tit", "Inspector", "--yar", "35000"],
        description="W O R K",
    )
    worker = deserialize(Worker, result, Context())
    assert worker == Worker(name="Gunther", job=Job(title="Inspector", yearly_pay=35000))

    result = parse_cli(
        root_type=Worker,
        argv=["--name", "Gunther", "--title", "Inspector", "--yearly-pay", "35000"],
    )
    worker = deserialize(Worker, result, Context())
    assert worker == Worker(name="Gunther", job=Job(title="Inspector", yearly_pay=35000))


@dataclass
class Konfig:
    mode: Literal["dev", "prod", "test"]
    level: Literal["debug", "info", "warning", "error"]


def test_parse_literal():
    result = parse_cli(
        root_type=Konfig,
        argv=["--mode", "dev", "--level", "debug"],
    )
    config = deserialize(Konfig, result, Context())
    assert config == Konfig(mode="dev", level="debug")

    result = parse_cli(
        root_type=Konfig,
        argv=["--mode", "prod", "--level", "info"],
    )
    config = deserialize(Konfig, result, Context())
    assert config == Konfig(mode="prod", level="info")

    with pytest.raises(SystemExit):
        parse_cli(
            root_type=Konfig,
            argv=["--mode", "invalid", "--level", "debug"],
        )

    with pytest.raises(SystemExit):
        parse_cli(
            root_type=Konfig,
            argv=["--mode", "dev", "--level", "invalid"],
        )


def test_subcommands_tagset():
    doables = TagDict(
        {
            "nom": Eat,
            "zzz": Sleep,
        }
    )

    def do(*args):
        result = deserialize(
            Any @ doables,
            CommandLineArguments(args),
            Context(),
        )
        return result.do()

    assert do("nom", "--food", "jam") == "I eat jam"
    assert do("zzz", "--hours", "8") == "I sleep 8 hours"


def add_them(x: int, y: int) -> int:
    return x + y


def mul_them(x: int, y: int) -> int:
    return x * y


def test_clargs_call():
    def do(*args):
        return deserialize(Call[add_them], CommandLineArguments(args))

    assert do("-x", "3", "-y", "4") == 7


def test_clargs_auto():
    def do(*args):
        return deserialize(Auto[add_them], CommandLineArguments(args))

    assert do("-x", "3", "-y", "4")() == 7


def test_clargs_auto_union():
    def do(*args):
        return deserialize(TaggedUnion[Call[add_them], Call[mul_them]], CommandLineArguments(args))

    assert do("add_them", "-x", "3", "-y", "4") == 7
    assert do("mul_them", "-x", "3", "-y", "4") == 12


def test_clargs_tagset():
    ts = TagDict({"+": Call[add_them], "*": Call[mul_them]})

    def do(*args):
        return deserialize(Annotated[int, ts], CommandLineArguments(args))

    assert do("+", "-x", "3", "-y", "4") == 7
    assert do("*", "-x", "3", "-y", "4") == 12



import pytest

from serieux import deserialize, schema, serialize
from serieux.exc import ValidationError
from serieux.features.registered import Registered, StringMapped, singleton


class Numero(StringMapped):
    one = 1
    two = 2
    three = 3


NumeroDeux = StringMapped.create("NumeroDeux", {"one": 1, "two": 2, "three": 3})


class Person(Registered):
    def __init__(self, name, age):
        super().__init__(name)
        self.name = name
        self.age = age


class SuperPerson(Person):
    def __init__(self, name, age, power):
        super().__init__(name, age)
        self.power = power


anita = Person("anita", 76)
bernard = Person("bernard", 73)
charlotte = SuperPerson("charlotte", 33, "spits fire")


def test_string_mapped_deserialize():
    assert deserialize(Numero, "one") == 1


def test_string_mapped_deserialize2():
    assert deserialize(NumeroDeux, "one") == 1


def test_string_mapped_deserialize_unknown():
    with pytest.raises(ValidationError, match="'bone' is not a registered option"):
        deserialize(Numero, "bone")


def test_string_mapped_serialize():
    assert serialize(Numero, 1) == "one"


def test_string_mapped_serialize_unknown():
    with pytest.raises(ValidationError, match="The value '33' is not registered"):
        serialize(Numero, 33)


def test_string_mapped_schema():
    sch = schema(Numero).compile(root=False)

    assert sch == {"type": "string", "enum": ["one", "two", "three"]}


def test_registered_serialize():
    assert serialize(Person, anita) == "anita"
    assert serialize(Person, bernard) == "bernard"
    assert serialize(Person, charlotte) == "charlotte"


def test_registered_deserialize():
    assert deserialize(Person, "anita") is anita
    assert deserialize(Person, "bernard") is bernard
    assert deserialize(Person, "charlotte") is charlotte


def test_registered_schema():
    sch = schema(Person).compile(root=False)

    assert sch == {"type": "string", "enum": ["anita", "bernard", "charlotte"]}


class Tool(Registered):
    pass


@singleton("hammer")
class Hammer(Tool):
    def use(self):
        return "bang!"


@singleton
class Saw(Tool):
    def use(self):
        return "zing!"


def test_singleton():
    assert deserialize(Tool, "hammer") is Hammer
    assert deserialize(Tool, "saw") is Saw


def test_illegal_singleton():
    with pytest.raises(
        TypeError, match="must be a subclass of Registered, but not a direct subclass"
    ):

        @singleton
        class Bloop(Registered):
            pass



from typing import Any

import pytest

from serieux import Serieux
from serieux.ctx import Context
from serieux.exc import ValidationError
from serieux.features.autotag import AutoTagAny
from serieux.features.tagset import tag_field, value_field

from ..definitions import Point

srx = (Serieux + AutoTagAny)()
deserialize = srx.deserialize
serialize = srx.serialize
schema = srx.schema


class Pointato:
    x: int
    y: int


def test_serialize_point():
    ctx = Context()
    obj = Point(13, 14)
    result = serialize(object, obj, ctx)

    assert result == {tag_field: f"{Point.__module__}:Point", "x": 13, "y": 14}


def test_serialize_point_any():
    ctx = Context()
    obj = Point(13, 14)
    result = serialize(Any, obj, ctx)

    assert result == {tag_field: f"{Point.__module__}:Point", "x": 13, "y": 14}


def test_serialize_nested_class():
    class Nested(Point):
        pass

    ctx = Context()
    obj = Nested(44, 55)

    with pytest.raises(ValidationError, match="Only top-level symbols can be serialized"):
        serialize(object, obj, ctx)


def test_deserialize_point():
    ctx = Context()
    data = {tag_field: f"{Point.__module__}:Point", "x": 13, "y": 14}

    result = deserialize(object, data, ctx)
    expected = Point(13, 14)
    assert result == expected


def test_deserialize_point_any():
    ctx = Context()
    data = {tag_field: f"{Point.__module__}:Point", "x": 13, "y": 14}

    result = deserialize(Any, data, ctx)
    expected = Point(13, 14)
    assert result == expected


def test_deserialize_not_hit_unless_object():
    ctx = Context()
    data = {tag_field: f"{Point.__module__}:Point", "x": 13, "y": 14}

    # Passing Pointato instead of object should not trigger
    # the codepath for AutoTagAny
    with pytest.raises(ValidationError):
        deserialize(Pointato, data, ctx)


def test_deserialize_without_class():
    ctx = Context()
    data = {"x": 13, "y": 14}

    with pytest.raises(ValidationError, match="No default class is defined"):
        deserialize(object, data, ctx)


def test_deserialize_vague_class():
    ctx = Context()
    data = {tag_field: "Point", "x": 13, "y": 14}
    with pytest.raises(ValidationError, match="no default module is defined"):
        deserialize(object, data, ctx)


def test_deserialize_invalid_class_reference():
    ctx = Context()
    data = {tag_field: "invalid:reference", "value": "test"}

    with pytest.raises(ValidationError, match="ModuleNotFoundError"):
        deserialize(object, data, ctx)


def test_deserialize_bad_class_format():
    ctx = Context()
    data = {tag_field: "too:many:colons", "value": "test"}

    with pytest.raises(ValidationError, match="Bad format for class reference"):
        deserialize(object, data, ctx)


def test_serialize_int():
    ctx = Context()
    result = serialize(Any, 46, ctx)

    assert result == {
        tag_field: "builtins:int",
        value_field: 46,
    }


def test_deserialize_int():
    ctx = Context()
    data = {tag_field: "builtins:int", value_field: 42}
    result = deserialize(Any, data, ctx)
    assert result == 42



from functools import cached_property


class ProxyBase:
    def __getattribute__(self, name):
        if name in ("_obj", "_computing", "_evaluate", "_type", "__dict__", "_"):
            return object.__getattribute__(self, name)
        elif name == "__class__":
            return object.__getattribute__(self, "_type") or type(self)
        return getattr(self._obj, name)

    def __str__(self):
        return str(self._obj)

    def __repr__(self):
        return repr(self._obj)

    def __eq__(self, other):
        return self._obj == other

    def __ne__(self, other):
        return self._obj != other

    def __lt__(self, other):
        return self._obj < other

    def __le__(self, other):
        return self._obj <= other

    def __gt__(self, other):
        return self._obj > other

    def __ge__(self, other):
        return self._obj >= other

    def __hash__(self):
        return hash(self._obj)

    def __len__(self):
        return len(self._obj)

    def __getitem__(self, key):
        return self._obj[key]

    def __iter__(self):
        return iter(self._obj)

    def __bool__(self):
        return bool(self._obj)

    def __contains__(self, item):
        return item in self._obj

    def __add__(self, other):
        return self._obj + other

    def __sub__(self, other):
        return self._obj - other

    def __mul__(self, other):
        return self._obj * other

    def __truediv__(self, other):
        return self._obj / other

    def __floordiv__(self, other):
        return self._obj // other

    def __mod__(self, other):
        return self._obj % other

    def __pow__(self, other):
        return self._obj**other

    def __radd__(self, other):
        return other + self._obj

    def __rsub__(self, other):
        return other - self._obj

    def __rmul__(self, other):
        return other * self._obj

    def __rtruediv__(self, other):
        return other / self._obj

    def __rfloordiv__(self, other):
        return other // self._obj

    def __rmod__(self, other):
        return other % self._obj

    def __rpow__(self, other):
        return other**self._obj

    def __neg__(self):
        return -self._obj

    def __pos__(self):
        return +self._obj

    def __abs__(self):
        return abs(self._obj)

    def __call__(self, *args, **kwargs):
        return self._obj(*args, **kwargs)


class LazyProxy(ProxyBase):
    def __init__(self, evaluate, type=None):
        self._type = type
        self._evaluate = evaluate
        self._computing = False

    @cached_property
    def _obj(self):
        if self._computing:
            raise Exception("Deadlock: asked for a value during its computation.")
        self._computing = True
        try:
            rval = self._evaluate()
            if isinstance(rval, LazyProxy):  # pragma: no cover
                return rval._obj
        finally:
            self._computing = False
        return rval


class CommentProxy(ProxyBase):
    def __init__(self, obj, comment):
        self._obj = obj
        self._type = type(obj)
        self._ = comment



import inspect

import pytest

from serieux import Serieux, deserialize, serialize
from serieux.ctx import Context

from .definitions import Defaults, Point, World
from .test_serialize import Special

SEP = """
======
"""


def getcodes(fn, *sigs):
    sigs = [(sig if isinstance(sig, tuple) else (sig,)) for sig in sigs]
    codes = [inspect.getsource(fn.resolve(*sig)) for sig in sigs]
    return SEP.join(codes)


@pytest.mark.parametrize("cls", [Point, World, Defaults])
def test_serialize_codegen(cls, file_regression):
    code = getcodes(serialize, (type[cls], cls, Context))
    file_regression.check(code)


@pytest.mark.parametrize("cls", [Point, World, Defaults])
def test_deserialize_codegen(cls, file_regression):
    code = getcodes(deserialize, (type[cls], dict, Context))
    file_regression.check(code)


def test_special_serializer_codegen(file_regression):
    custom = (Serieux + Special)()
    code = getcodes(custom.serialize, (type[Point], Point, Context))
    file_regression.check(code)



from typing import Any

from ovld import Medley, call_next, ovld

from ..ctx import Context
from ..model import FieldModelizable
from ..priority import HI2


def unflatten(d: dict):
    rval = {}
    split_keys = [(k.split("."), v) for k, v in d.items()]
    for parts, v in sorted(split_keys, key=lambda kv: len(kv[0])):
        current = rval
        for p in parts[:-1]:
            current = current.setdefault(p, {})
        current[parts[-1]] = v
    return rval


class DottedNotation(Medley):
    @ovld(priority=HI2)
    def deserialize(self, t: Any, obj: dict, ctx: Context):
        if issubclass(t, FieldModelizable) and any("." in k for k in obj.keys()):
            return call_next(t, unflatten(obj), ctx)
        return call_next(t, obj, ctx)



from serieux import Serieux
from serieux.ctx import Context
from serieux.features.dotted import DottedNotation, unflatten
from tests.common import one_test_per_assert
from tests.definitions import Job, Worker

deserialize = (Serieux + DottedNotation)().deserialize


def test_unflatten():
    assert unflatten({"x.a": 1, "x.b": 2, "y": 3}) == {"x": {"a": 1, "b": 2}, "y": 3}


def test_unflatten_deep():
    assert unflatten({"a.b.c.d.e": 8}) == {"a": {"b": {"c": {"d": {"e": 8}}}}}


@one_test_per_assert
def test_unflatten_merge():
    assert unflatten({"x": {"a": 1}, "x.b": 2, "y": 3}) == {"x": {"a": 1, "b": 2}, "y": 3}
    assert unflatten({"x.b": 2, "y": 3, "x": {"a": 1}}) == {"x": {"a": 1, "b": 2}, "y": 3}
    assert unflatten({"a.b.c": {"z": 99}, "a.b.c.d.e": 8}) == {
        "a": {"b": {"c": {"z": 99, "d": {"e": 8}}}}
    }


def test_unflatten_ordering():
    d1 = {"b": 1, "a": 2, "r": 3, "n": 4}
    d2 = unflatten(d1)
    assert list(d2) == list("barn")

    d1 = {"b": 1, "a.x": 2, "r": 3, "n": 4}
    d2 = unflatten(d1)
    assert list(d2) == list("brna")

    d1 = {"b": 1, "a": {"y": 7}, "a.x": 2, "r": 3, "n": 4}
    d2 = unflatten(d1)
    assert list(d2) == list("barn")


def test_deserialize():
    data = {"name": "John Doe", "job.title": "Software Engineer", "job.yearly_pay": 80000.0}
    assert deserialize(Worker, data, Context()) == Worker(
        name="John Doe", job=Job(title="Software Engineer", yearly_pay=80000.0)
    )



from dataclasses import replace
from typing import Any

from ovld import Medley, call_next, ovld, recurse
from ovld.types import HasMethod
from ovld.utils import ResolutionError

from ..ctx import Context
from ..model import Model, model
from ..priority import STD4


class UserMethods(Medley):
    @ovld(priority=STD4)
    def deserialize(self, t: type[HasMethod["serieux_deserialize"]], obj: Any, ctx: Context):  # noqa: F821
        def cn(t, obj, ctx, *, from_top=False):
            return recurse(t, obj, ctx) if from_top else call_next(t, obj, ctx)

        try:
            return t.serieux_deserialize(obj, ctx, cn)
        except ResolutionError:
            # If t implements serieux_deserialize with ovld and no method matches, it will
            # throw a ResolutionError and we simply resume our search down the stack.
            return call_next(t, obj, ctx)

    @ovld(priority=STD4)
    def serialize(self, t: type[HasMethod["serieux_serialize"]], obj: Any, ctx: Context):  # noqa: F821
        def cn(t, obj, ctx, *, from_top=False):
            return recurse(t, obj, ctx) if from_top else call_next(t, obj, ctx)

        try:
            return t.serieux_serialize(obj, ctx, cn)
        except ResolutionError:
            return call_next(t, obj, ctx)

    @ovld(priority=STD4)
    def schema(self, t: type[HasMethod["serieux_schema"]], ctx: Context):  # noqa: F821
        def cn(t, ctx, *, from_top=False):
            return recurse(t, ctx) if from_top else call_next(t, ctx)

        try:
            return t.serieux_schema(ctx, cn)
        except ResolutionError:  # pragma: no cover
            return call_next(t, ctx)


@model.register(priority=1)
def _(t: type[HasMethod["serieux_model"]]):  # noqa: F821
    def cn(t, *, from_top=False):  # pragma: no cover
        return recurse(t) if from_top else call_next(t)

    return t.serieux_model(cn)


@model.register(priority=2)
def _(t: type[HasMethod["serieux_to_string"]] | type[HasMethod["serieux_from_string"]]):  # noqa: F821
    m = call_next(t)
    if not m:
        m = Model(t, fields=None)
    if hasattr(t, "serieux_to_string"):
        m = replace(m, to_string=t.serieux_to_string)
    if hasattr(t, "serieux_from_string"):
        m = replace(m, from_string=t.serieux_from_string)
    return m



import json
import sys
from contextlib import contextmanager
from io import StringIO
from pathlib import Path

import pytest

from serieux import Serieux
from serieux.exc import SerieuxError

datapath = Path(__file__).parent / "data"


@pytest.hookimpl()
def pytest_exception_interact(node, call, report):
    if issubclass(call.excinfo.type, SerieuxError):
        exc = call.excinfo.value
        io = StringIO()
        exc.display(file=io)
        entry = report.longrepr.reprtraceback.reprentries[-1]
        entry.style = "short"
        content = io.getvalue()
        entry.lines = [content] + [""] * content.count("\n")
        report.longrepr.reprtraceback.reprentries = [entry]


@pytest.fixture
def check_error_display(capsys, file_regression):
    @contextmanager
    def check(message=None, exc_type=SerieuxError):
        with pytest.raises(exc_type, match=message) as exc:
            yield

        exc.value.display(file=sys.stderr)
        cap = capsys.readouterr()
        out = cap.out.replace(str(datapath.parent), "REDACTED")
        err = cap.err.replace(str(datapath.parent), "REDACTED")
        file_regression.check("\n".join([out, "=" * 80, err]))

    yield check


@pytest.fixture
def fresh_serieux(monkeypatch):
    """Fixture that monkeypatches serieux module components for testing."""

    class NewSerieux(Serieux):
        pass

    new_serieux = NewSerieux()

    monkeypatch.setattr("serieux.serieux", new_serieux)
    monkeypatch.setattr("serieux.Serieux", NewSerieux)
    monkeypatch.setattr("serieux.deserialize", new_serieux.deserialize)
    monkeypatch.setattr("serieux.serialize", new_serieux.serialize)
    return new_serieux


@pytest.fixture
def schematest(file_regression):
    ABSENT = object()

    def check(type, value=ABSENT):
        import jsonschema

        from serieux import schema, serialize

        sch = schema(type)
        sch_value = sch.compile()
        if value is not ABSENT:
            jsonschema.validate(serialize(type, value), sch_value)

        file_regression.check(json.dumps(sch_value, indent=4))

    yield check



import argparse
import sys
from dataclasses import dataclass
from functools import cache

from ovld import Medley, recurse
from ovld.dependent import Regexp
from rich.pretty import pprint

from serieux import Serieux, Sources, deserialize
from serieux.ctx import Context
from serieux.instructions import strip
from serieux.model import FieldModelizable, model

##################
# Implementation #
##################


@dataclass
class CommandLineArguments:
    arguments: list[str]


@cache
def parser_for(t: type):
    m = model(t)
    parser = argparse.ArgumentParser(description=f"Arguments for {m.original_type.__name__}")
    for field in m.fields:
        if not field.name.startswith("_"):
            typ = strip(field.type)
            if typ not in (int, float, str, bool):
                typ = str
            parser.add_argument(
                f"--{field.name}",
                type=typ,
                help=field.description or field.name,
                required=field.required,
            )
    return parser


@Serieux.extend
class FromArguments(Medley):
    def deserialize(self, t: type[FieldModelizable], obj: CommandLineArguments, ctx: Context):
        parser = parser_for(t)
        ns = parser.parse_args(obj.arguments)
        values = {k: v for k, v in vars(ns).items() if v is not None}
        return recurse(t, values, ctx)


#################
# Demonstration #
#################


@dataclass
class RGB:
    red: int
    green: int
    blue: int


@Serieux.extend
class RGBSerializer(Medley):
    def deserialize(self, t: type[RGB], obj: Regexp[r"^#[0-9a-fA-F]{6}$"], ctx: Context):
        hex_str = obj.lstrip("#")
        red = int(hex_str[0:2], 16)
        green = int(hex_str[2:4], 16)
        blue = int(hex_str[4:6], 16)
        return RGB(red=red, green=green, blue=blue)


@dataclass
class Person:
    # Name of the person
    name: str
    # Age of the person
    age: int
    # The person's shirt color
    shirt: RGB


def main(argv=["--name", "Travis", "--shirt", "#ff0010"]):
    defaults = {"name": "Guy", "age": 0, "shirt": "#ffffff"}
    overrides = {"shirt": {"blue": 128}}
    person = deserialize(Person, Sources(defaults, CommandLineArguments(argv), overrides))
    pprint(person, expand_all=True)


if __name__ == "__main__":
    main(sys.argv[1:])



import json
import os
import re
from dataclasses import field
from pathlib import Path
from types import NoneType
from typing import Any, Literal, get_args

from ovld import Medley, call_next, ovld, recurse
from ovld.dependent import Regexp

from ..ctx import AccessPath
from ..exc import NotGivenError, ValidationError
from ..instructions import strip
from ..priority import HI1
from ..utils import UnionAlias
from .lazy import LazyProxy
from .partial import Sources


@ovld
def decode_string(t: type[int] | type[float] | type[str], value: str):
    return t(value)


@ovld
def decode_string(t: type[NoneType], value: str):
    val = value.lower()
    if val in ("", "null", "none"):
        return None
    else:
        raise ValidationError(f"Cannot convert '{value}' to None")


@ovld
def decode_string(t: type[bool], value: str):
    val = value.lower()
    if val in ("true", "1", "yes", "on"):
        return True
    elif val in ("false", "0", "no", "off", ""):
        return False
    else:
        raise ValidationError(f"Cannot convert '{value}' to boolean")


@ovld
def decode_string(t: type[object], value: str):
    try:
        return json.loads(value)
    except json.JSONDecodeError:
        return value


@ovld
def decode_string(t: type[UnionAlias], value: str):
    err = None
    for opt in get_args(t):
        try:
            return decode_string(opt, value)
        except Exception as exc:
            err = exc
    raise err


@ovld
def decode_string(t: type[list], value: str):
    (element_type,) = get_args(t) or (object,)
    return [recurse(element_type, item.strip()) for item in str(value).split(",")]


class Environment(AccessPath):
    refs: dict[tuple[str, ...], object] = field(default_factory=dict, repr=False)
    environ: dict = field(default_factory=lambda: os.environ, repr=False)

    def evaluate_reference(self, ref):
        def try_int(x):
            try:
                return int(x)
            except ValueError:
                return x

        stripped = ref.lstrip(".")
        dots = len(ref) - len(stripped)
        root = () if not dots else self.access_path[:-dots]
        parts = [try_int(x) for x in stripped.split(".")]
        return self.refs[(*root, *parts)]

    @ovld
    def resolve_variable(self, t: Any, expr: str, /):
        match expr.split(":", 1):
            case (method, expr):
                return recurse(t, method, expr)
            case _:
                return recurse(t, "", expr)

    def resolve_variable(self, t: Any, method: Literal[""], expr: str, /):
        return LazyProxy(lambda: self.evaluate_reference(expr))

    def resolve_variable(self, t: Any, method: Literal["env"], expr: str, /):
        try:
            env_value = self.environ[expr]
        except KeyError:
            raise NotGivenError(f"Environment variable '{expr}' is not defined")
        else:
            return decode_string(t, env_value)

    def resolve_variable(self, t: Any, method: Literal["envfile"], expr: str, /):
        try:
            pth = Path(self.environ[expr]).expanduser()
        except KeyError:
            raise NotGivenError(f"Environment variable '{expr}' is not defined")
        if pth.exists():
            return pth
        else:
            return Sources(*[Path(x.strip()).expanduser() for x in str(pth).split(",")])

    def resolve_variable(self, t: Any, method: str, expr: str, /):
        raise ValidationError(
            f"Cannot resolve '{method}:{expr}' because the '{method}' resolver is not defined."
        )


class Interpolation(Medley):
    @ovld(priority=HI1(3))
    def deserialize(self, t: Any, obj: object, ctx: Environment):
        rval = call_next(t, obj, ctx)
        ctx.refs[ctx.access_path] = rval
        return rval

    @ovld(priority=HI1(2))
    def deserialize(self, t: Any, obj: Regexp[r"^\$\{[^}]+\}$"], ctx: Environment):
        expr = obj.lstrip("${").rstrip("}")
        obj = ctx.resolve_variable(strip(t), expr)
        if isinstance(obj, LazyProxy):

            def interpolate():
                return recurse(t, obj._obj, ctx)

            return LazyProxy(interpolate)
        else:
            return recurse(t, obj, ctx)

    @ovld(priority=HI1(1))
    def deserialize(self, t: Any, obj: Regexp[r"\$\{[^}]+\}"], ctx: Environment):
        def interpolate():
            def repl(match):
                return str(ctx.resolve_variable(str, match.group(1)))

            subbed = re.sub(r"\$\{([^}]+)\}", repl, obj)
            return recurse(t, subbed, ctx)

        return LazyProxy(interpolate)



from dataclasses import dataclass

from ovld import Medley, call_next, ovld, recurse
from rich.pretty import pprint

from serieux import Context, Serieux, Sources, deserialize

##################
# Implementation #
##################


@Serieux.extend
class Dotted(Medley):
    @ovld(priority=10)
    def deserialize(self, t: type[object], obj: dict, ctx: Context):
        if any("." in k for k in obj.keys()):
            parts = []
            for k, v in obj.items():
                levels = k.split(".")
                current = v
                while levels:
                    current = {levels.pop(): current}
                parts.append(current)
            return recurse(t, Sources(*parts), ctx)
        return call_next(t, obj, ctx)


#################
# Demonstration #
#################


@dataclass
class Climate:
    hot: bool
    sunny: bool


@dataclass
class City:
    name: str
    population: int
    climate: Climate


@dataclass
class Country:
    name: str
    capital: City


def main():
    data = {
        "name": "Canada",
        "capital": {
            "name": "Ottawa",
        },
        "capital.population": 800_000,
        "capital.climate.hot": False,
        "capital.climate.sunny": False,
    }
    print("\n== Serialized ==\n")
    pprint(data)
    deser = deserialize(Country, data)
    print("\n== Deserialized ==\n")
    pprint(deser)
    assert deser == Country(
        name="Canada",
        capital=City(
            name="Ottawa",
            population=800_000,
            climate=Climate(
                hot=False,
                sunny=False,
            ),
        ),
    )


if __name__ == "__main__":
    main()



import importlib
import importlib.metadata
from collections import deque
from dataclasses import dataclass, field
from functools import cached_property
from typing import TYPE_CHECKING, Annotated, Any, Callable, Iterable, TypeAlias, Union

from ovld import Medley, call_next, ovld, recurse

from ..ctx import Context
from ..exc import ValidationError
from ..instructions import BaseInstruction, Instruction, T, annotate, pushdown, strip
from ..model import constructed_type
from ..schema import AnnotatedSchema
from ..tell import KeyValueTell, TypeTell, tells

tag_field = "$class"
value_field = "$value"


class TagSet(BaseInstruction):  # pragma: no cover
    def get_type(self, tag: str | None, ctx: Context = None) -> type:
        raise NotImplementedError()

    def get_tag(self, t: type, ctx: Context = None) -> str | None:
        raise NotImplementedError()

    def iterate(self, base: type, ctx: Context = None) -> Iterable[tuple[str | None, type]]:
        raise NotImplementedError()

    def closed(self, base):
        return True


@dataclass(frozen=True, eq=False)
class TagDict(TagSet):
    possibilities: dict = field(default_factory=dict)

    def register(self, tag_or_cls=None, cls=None):
        if isinstance(tag_or_cls, str):
            tag = tag_or_cls

            def decorator(cls):
                self.possibilities[tag] = cls
                return cls

            return decorator if cls is None else decorator(cls)

        else:
            assert cls is None
            cls = tag_or_cls
            tag = cls.__name__.lower()
            self.possibilities[tag] = cls
            return cls

    def get_type(self, tag: str | None, ctx: Context = None) -> type:
        if tag is None:
            return self.get_type("default", ctx)
        try:
            return self.possibilities[tag]
        except KeyError:
            raise ValidationError(f"Tag '{tag}' is not registered", ctx=ctx)

    def get_tag(self, t: type, ctx: Context = None) -> str | None:
        for tag, cls in self.possibilities.items():
            if cls is t:
                return None if tag == "default" else tag
        raise ValidationError(f"No tag is registered for type '{t}'", ctx=ctx)

    def iterate(self, base: type, ctx: Context = None) -> Iterable[tuple[str | None, type]]:
        yield from self.possibilities.items()


@dataclass(frozen=True)
class Tag(TagSet):
    tag: str
    cls: type

    def get_type(self, tag: str | None, ctx: Context = None) -> type:
        if tag is None:
            raise ValidationError(f"Tag '{self.tag}' is required", ctx=ctx)
        if tag == self.tag:
            return self.cls
        raise ValidationError(f"Tag '{tag}' does not match expected tag '{self.tag}'", ctx=ctx)

    def get_tag(self, t: type, ctx: Context = None) -> str | None:
        if t is self.cls:
            return self.tag
        raise ValidationError(f"Type '{t}' does not match expected class '{self.cls}'", ctx=ctx)

    def iterate(self, base: type, ctx: Context = None) -> Iterable[tuple[str | None, type]]:
        if isinstance(base, type):
            assert issubclass(self.cls, base)
        yield (self.tag, self.cls)


@dataclass(frozen=True)
class FromEntryPoint(TagSet):
    entry_point: str
    default: type = None
    wrap: Callable = None

    @cached_property
    def elements(self):
        def _wrap(t):
            match self.wrap:
                case None:
                    return t
                case type() | Instruction():  # pragma: no cover
                    return self.wrap[t]
                case _:
                    return self.wrap(t)

        eps = importlib.metadata.entry_points(group=self.entry_point)
        return {ep.name: cls for ep in eps if (cls := _wrap(ep.load())) is not None}

    def get_type(self, tag: str | None, ctx: Context) -> type:
        eps = self.elements
        if tag is None:
            if self.default is not None:
                return self.default
            raise ValidationError("No tag provided for entry point lookup", ctx=ctx)
        try:
            return eps[tag]
        except KeyError:
            raise ValidationError(
                f"Tag '{tag}' is not registered in entry point group '{self.entry_point}'", ctx=ctx
            )

    def get_tag(self, t: type, ctx: Context) -> str | None:
        if t is self.default:
            return None
        for name, cls in self.elements.items():
            if cls is t:
                return name
        raise ValidationError(
            f"No entry point tag is registered for type '{t}' in group '{self.entry_point}'",
            ctx=ctx,
        )

    def iterate(self, base: type, ctx: Context = None) -> Iterable[tuple[str | None, type]]:
        for name, cls in self.elements.items():
            if base is Any or issubclass(cls, base):
                yield (name, cls)
        if self.default is not None and (base is Any or issubclass(self.default, base)):
            yield (None, self.default)


@dataclass(frozen=True)
class ReferencedClass(TagSet):
    default: type = None
    default_module: str = None

    def get_type(self, tag: str | None, ctx: Context) -> type:
        if tag is None:
            if self.default is not None:
                return self.default
            else:
                raise ValidationError("No default class is defined when there is no explicit tag")

        if (ncolon := tag.count(":")) == 0:
            mod_name = self.default_module
            if mod_name is None:
                raise ValidationError(
                    "The reference does not specify a module and no default module is defined",
                    ctx=ctx,
                )
            symbol = tag
        elif ncolon == 1:
            mod_name, symbol = tag.split(":")
        else:
            raise ValidationError(f"Bad format for class reference: '{tag}'", ctx=ctx)
        try:
            mod = importlib.import_module(mod_name)
            return getattr(mod, symbol)
        except (ModuleNotFoundError, AttributeError) as exc:
            raise ValidationError(exc=exc, ctx=ctx)

    def get_tag(self, t: type, ctx: Context) -> str | None:
        qn = t.__qualname__
        if "." in qn:
            raise ValidationError("Only top-level symbols can be serialized", ctx=ctx)
        mod = t.__module__
        return f"{mod}:{qn}"

    def iterate(self, base: type, ctx: Context = None) -> Iterable[tuple[str | None, type]]:
        if base is Any or base is object:
            return
        queue = deque([base])
        while queue:
            sc = queue.popleft()
            sc_mod = sc.__module__
            sc_name = sc.__name__
            if sc is self.default:
                tag = None
            elif sc_mod is self.default_module:
                tag = sc_name
            else:
                tag = f"{sc_mod}:{sc_name}"
            yield tag, sc
            queue.extend(sc.__subclasses__())

    def closed(self, base):
        return base is not Any and base is not object

    def __call__(self, *args, **kwargs):
        return type(self)(*args, **kwargs)


Referenced = ReferencedClass()


class MultiTagSet(TagSet):
    def __init__(self, *tagsets):
        assert tagsets
        self.tagsets = tagsets

    def get_type(self, tag, ctx):
        for ts in self.tagsets:
            try:
                return ts.get_type(tag, ctx)
            except ValidationError:
                pass
        raise ValidationError("No tagset could resolve the tag", ctx=ctx)

    def get_tag(self, t, ctx):
        for ts in self.tagsets:
            try:
                return ts.get_tag(t, ctx)
            except ValidationError:
                pass
        raise ValidationError(f"No tagset could resolve for type {t}", ctx=ctx)

    def iterate(self, base, ctx=None):
        seen = set()
        for ts in self.tagsets:
            for tag, sc in ts.iterate(base, ctx):
                if tag not in seen:
                    seen.add(tag)
                    yield tag, sc

    def closed(self, base):
        return all(ts.closed(base) for ts in self.tagsets)


def decompose(annt):
    base = pushdown(annt)
    match list(TagSet.extract_all(annt)):
        case (ts,):
            pass
        case many:
            ts = MultiTagSet(*many)
    return base, ts


class TagSetFeature(Medley):
    @ovld(priority=10)
    def serialize(self, t: type[Any @ TagSet], obj: object, ctx: Context, /):
        base, ts = decompose(t)
        if base is not Any and not isinstance(obj, base):
            raise ValidationError(f"'{obj}' is not a subclass of '{base}'", ctx=ctx)
        objt = type(obj)
        tag = ts.get_tag(objt, ctx)
        rval = call_next(objt, obj, ctx)
        if not isinstance(rval, dict):
            rval = {value_field: rval}
        if tag is not None:
            rval[tag_field] = tag
        return rval

    def deserialize(self, t: type[Any @ TagSet], obj: dict, ctx: Context, /):
        base, ts = decompose(t)
        obj = dict(obj)
        tag = obj.pop(tag_field, None)
        obj = obj.pop(value_field, obj)
        if tag is not None:
            tag = recurse(str, tag, ctx)
        declared = ts.get_type(tag, ctx)
        if base is not Any and base is not object and isinstance(base, type):
            actual_class = constructed_type(declared)
            if not issubclass(actual_class, base):
                raise ValidationError(f"'{actual_class}' is not a subclass of '{base}'", ctx=ctx)
        return recurse(strip(annotate(declared, t), TagSet), obj, ctx)

    def schema(self, t: type[Any @ TagSet], ctx: Context):
        base, ts = decompose(t)
        subschemas = []
        for tag, sc in ts.iterate(base, ctx):
            if base is not Any and not issubclass(sc, base):  # pragma: no cover
                continue
            subsch = recurse(strip(annotate(sc, t), TagSet))
            if tag is not None:
                subsch = AnnotatedSchema(
                    parent=subsch,
                    properties={
                        tag_field: {
                            "description": "Reference to the class to instantiate",
                            "const": tag,
                        }
                    },
                    required=[tag_field],
                )
            subschemas.append(subsch)
        if not ts.closed(base):
            subschemas.append({"type": "object", "additionalProperties": True})
        if len(subschemas) == 1:
            return subschemas[0]
        else:
            return {"oneOf": subschemas}


@tells.register(priority=1)
def tells(typ: type[Any @ TagSet]):
    base, ts = decompose(typ)
    kvt = [KeyValueTell(tag_field, tag) for tag, _ in ts.iterate(base)]
    return {TypeTell(dict), *kvt}


if TYPE_CHECKING:
    TaggedSubclass: TypeAlias = Annotated[T, None]
    Tagged: TypeAlias = Annotated
    TaggedUnion = Union

else:

    class TaggedSubclass:
        def __class_getitem__(cls, item):
            return Annotated[item, Referenced(default=item, default_module=item.__module__)]

    class Tagged(type):
        def __class_getitem__(cls, arg):
            match arg:
                case (t, name):
                    return Annotated[t, Tag(name, t)]
                case t:
                    st = strip(t)
                    tag = getattr(st, "serieux_tag", None) or st.__name__.lower()
                    return Annotated[t, Tag(tag, t)]

    class TaggedUnion(type):
        def __class_getitem__(cls, args):
            if isinstance(args, dict):
                return Union[tuple(Tagged[v, k] for k, v in args.items())]
            elif not isinstance(args, (list, tuple)):
                return Tagged[args]
            return Union[tuple(Tagged[arg] for arg in args)]



import importlib
import inspect
import typing
from dataclasses import MISSING, dataclass, replace
from functools import cached_property, partial
from typing import Annotated, Any

from ovld import call_next

from .docstrings import get_variable_data
from .instructions import BaseInstruction, inherit, strip
from .model import Field, Model, model
from .utils import evaluate_hint


class MeldedCall:
    def __init__(self, *funcs):
        self.steps = [
            (
                func,
                (params := list(inspect.signature(func).parameters.values())[min(i, 1) :]),
                set(p.name for p in params),
            )
            for i, func in enumerate(funcs)
        ]

    def __call__(self, **kwargs):
        current = inspect._empty
        for fn, _, argnames in self.steps:
            args = {an: kwargs[an] for an in argnames if an in kwargs}
            if current is inspect._empty:
                current = fn(**args)
            else:
                current = fn(current, **args)
        return current

    @cached_property
    def __variable_data__(self):
        vd = {}
        for fn, _, _ in self.steps:
            vd.update(get_variable_data(fn))
        return vd

    @cached_property
    def __signature__(self):
        final_params = []
        seen = set()
        for _, params, _ in self.steps:
            for param in params:
                if param.name not in seen:
                    seen.add(param.name)
                    new_param = param.replace(kind=inspect.Parameter.KEYWORD_ONLY)
                    final_params.append(new_param)
        return inspect.Signature(final_params)


@dataclass(frozen=True)
class Auto(BaseInstruction):
    call: bool = False
    embed_self: bool = True

    @property
    def annotation_priority(self):  # pragma: no cover
        return 1

    def __class_getitem__(cls, t):
        return cls()[t]

    def __getitem__(self, t):
        return Annotated[t, self]

    __call__ = replace


Call = Auto(call=True)


def model_from_callable(t, call=False, embed_self=True):
    orig_t, t = t, strip(t)
    if t is Any:
        return None
    if isinstance(t, type) and call:
        raise TypeError("Call[...] should only wrap callables")
    sig = inspect.signature(t)
    fields = []
    docs = get_variable_data(t)
    for param in sig.parameters.values():
        if param.name == "self" and param.annotation in (inspect._empty, typing.Self):
            parent_class = getattr(
                importlib.import_module(t.__module__), t.__qualname__.split(".")[0]
            )
            if embed_self:
                return model_from_callable(
                    MeldedCall(parent_class, t), call=call, embed_self=False
                )
            else:
                param = param.replace(annotation=parent_class)
        if param.annotation is inspect._empty:
            raise TypeError(f"Cannot model {t}: '{param.name}' lacks a type annotation.")
        field = Field(
            name=param.name,
            description=(docs[param.name].doc or param.name) if param.name in docs else param.name,
            metadata=(docs[param.name].metadata or {}) if param.name in docs else {},
            type=inherit(orig_t, evaluate_hint(param.annotation, None, None, None)),
            default=MISSING if param.default is inspect._empty else param.default,
            argument_name=param.name,
            property_name=None,
        )
        fields.append(field)

    if not isinstance(t, type) and not call:

        def build(*args, **kwargs):
            return partial(t, *args, **kwargs)

    else:
        build = t

    return Model(
        original_type=t,
        fields=fields,
        constructor=build,
    )


@model.register(priority=-1)
def _(t: type[Any @ Auto]):
    _, aut = Auto.decompose(t)
    aut = aut or Auto()
    if not aut.call and (normal := call_next(t)) is not None:
        return normal
    return model_from_callable(t, call=aut.call, embed_self=aut.embed_self)



import json
import logging
from collections import defaultdict
from dataclasses import dataclass, field, replace
from pathlib import Path
from typing import Any, Callable

from ovld.medley import ChainAll, KeepLast, Medley

logger = logging.getLogger(__name__)


class Context(Medley, default_combiner=KeepLast):
    follow = ChainAll()


class EmptyContext(Context):
    pass


class AccessPath(Context):
    full_path: tuple = ()

    @property
    def access_path(self):
        return tuple(k for _, _, k in self.full_path)

    def follow(self, objt, obj, field):
        return replace(self, full_path=(*self.full_path, (objt, obj, field)))


@dataclass
class Location:
    source: Path
    code: str
    start: int
    end: int
    linecols: tuple

    @property
    def text(self):  # pragma: no cover
        return self.code[self.start : self.end]


class Located(Context):
    location: Location = None


@dataclass
class Patch:
    data: Callable | Any
    ctx: Context = None
    description: str | None = None

    def __post_init__(self):
        if self.description is None:
            self.description = f"Set to: {self.data!r}"

    def compute(self):
        if callable(self.data):  # pragma: no cover
            return self.data()
        else:
            return self.data

    def __str__(self):  # pragma: no cover
        return f"Patch({self.description!r})"


class Patcher(Context):
    patches: dict[int, tuple[Context, Any]] = field(default_factory=dict)

    def declare_patch(self, patch):
        if not isinstance(patch, Patch):
            patch = Patch(patch, ctx=self)
        elif not patch.ctx:  # pragma: no cover
            patch = replace(patch, ctx=self)
        start = patch.ctx.location.start if isinstance(patch.ctx, Located) else None
        self.patches[start] = patch

    def apply_patches(self):
        codes = {}
        patches = defaultdict(list)
        for patch in self.patches.values():
            match patch.ctx:
                case Located(location=loc):
                    codes[loc.source] = loc.code
                    patches[loc.source].append((loc.start, loc.end, json.dumps(patch.compute())))
                case _:  # pragma: no cover
                    logger.warning(
                        f"Cannot apply patch at a context without a location: `{patch}`"
                    )

        for file, blocks in patches.items():
            code = codes[file].strip("\0")
            for start, end, content in sorted(blocks, reverse=True):
                code = code[:start] + content + code[end:]
            file.write_text(code)


empty = EmptyContext()



from typing import Any

from ovld import Medley, ovld

from ..ctx import Context
from ..exc import ValidationError
from ..priority import HI2


class RegisteredMC(type):
    def __init__(cls, name, bases, ns):
        super().__init__(name, bases, ns)
        if not hasattr(cls, "_registered_mapping") and bases != (SMBase,):
            cls._registered_mapping = {}
            cls._registered_inverse_mapping = {}


class SMBase:
    @staticmethod
    def create(name, mapping):
        inv_mapping = {v: k for k, v in mapping.items()}
        return RegisteredMC(
            name,
            (Registered,),
            {"_registered_mapping": mapping, "_registered_inverse_mapping": inv_mapping},
        )


class Registered(SMBase, metaclass=RegisteredMC):
    def __init__(self, registered_name):
        self.registered_name = registered_name
        type(self)._registered_mapping[registered_name] = self
        type(self)._registered_inverse_mapping[self] = registered_name


def singleton(arg, /):
    def wrap(cls):
        if (
            not isinstance(cls, type)
            or not issubclass(cls, Registered)
            or Registered in cls.__bases__
        ):
            raise TypeError(
                "@singleton must wrap a class definition and the class must be"
                " a subclass of Registered, but not a direct subclass."
            )
        return cls(registered_name=registered_name)

    if isinstance(arg, str):
        registered_name = arg
        return wrap
    else:
        registered_name = arg.__name__.lower()
        return wrap(arg)


class StringMappedMC(RegisteredMC):
    def __init__(cls, name, bases, ns):
        regs = {k: v for k, v in ns.items() if not k.startswith("__")}
        super().__init__(name, bases, ns)
        if regs:
            cls._registered_mapping.update(regs)
            cls._registered_inverse_mapping.update({v: k for k, v in regs.items()})


class StringMapped(SMBase, metaclass=StringMappedMC):
    pass


class RegisteredHandler(Medley):
    @ovld(priority=HI2)
    def deserialize(self, t: type[SMBase], obj: str, ctx: Context, /):
        mapping = t._registered_mapping
        if obj in mapping:
            return mapping[obj]
        else:
            raise ValidationError(
                f"'{obj}' is not a registered option. Should be one of: {list(mapping.keys())}"
            )

    @ovld(priority=HI2)
    def serialize(self, t: type[SMBase], obj: Any, ctx: Context, /):
        inv_mapping = t._registered_inverse_mapping
        if obj in inv_mapping:
            return inv_mapping[obj]
        else:
            raise ValidationError(
                f"The value '{obj}' is not registered under any name in the mapping"
            )

    @ovld(priority=HI2)
    def schema(self, t: type[SMBase], ctx: Context, /):
        return {"type": "string", "enum": list(t._registered_mapping.keys())}



from dataclasses import dataclass

import pytest

from serieux import Serieux, deserialize
from serieux.auto import Auto
from serieux.ctx import Context
from serieux.exc import ValidationError
from serieux.features.tagset import (
    Tag,
    Tagged,
    TaggedUnion,
    TagSet,
    TagSetFeature,
    tag_field,
    value_field,
)

from ..definitions import Player, Point

tu_serieux = (Serieux + TagSetFeature)()

deserialize = tu_serieux.deserialize
serialize = tu_serieux.serialize


def test_tagged_serialize():
    data = {tag_field: "point", "x": 1, "y": 2}
    assert serialize(Tagged[Point, "point"], Point(1, 2)) == data


def test_tagged_serialize_primitive():
    data = {tag_field: "nombre", value_field: 7}
    assert serialize(Tagged[int, "nombre"], 7) == data


def test_tagged_deserialize():
    data = {tag_field: "point", "x": 1, "y": 2}
    assert deserialize(Tagged[Point, "point"], data) == Point(1, 2)


def test_tagged_deserialize_primitive():
    data = {tag_field: "nombre", value_field: 7}
    assert deserialize(Tagged[int, "nombre"], data) == 7


def test_tunion_serialize():
    U = Tagged[Player, "player"] | Tagged[Point, "point"]
    data = {tag_field: "point", "x": 1, "y": 2}
    assert serialize(U, Point(1, 2), Context()) == data


def test_tunion_deserialize():
    U = Tagged[Player, "player"] | Tagged[Point, "point"]
    data = {tag_field: "point", "x": 1, "y": 2}
    assert deserialize(U, data) == Point(1, 2)


def test_tagged_default_tag():
    def f():
        pass

    assert TagSet.extract(Tagged[Point]).tag == "point"
    assert TagSet.extract(Tagged[Auto[f]]).tag == "f"


def test_tagged_union():
    us = [
        TaggedUnion[{"player": Player, "point": Point}],
        TaggedUnion[Player, Point],
        TaggedUnion[Point],
    ]
    for U in us:
        pt = Point(1, 2)
        data_point = {tag_field: "point", "x": 1, "y": 2}
        assert serialize(U, pt, Context()) == data_point
        assert deserialize(U, data_point) == pt

    for U in us[:-1]:
        ply = Player("Alice", "Smith", 0.333)
        data_player = {tag_field: "player", "first": "Alice", "last": "Smith", "batting": 0.333}
        assert serialize(U, ply, Context()) == data_player
        assert deserialize(U, data_player) == ply


@dataclass
class Blonde:
    name: str
    age: int


@dataclass
class Redhead:
    name: str
    age: int


def test_tagged_union_identical_fields():
    U = TaggedUnion[Blonde, Redhead]

    blonde = Blonde("Sam", 25)
    redhead = Redhead("Jack", 50)

    data_blonde = {tag_field: "blonde", "name": "Sam", "age": 25}
    data_redhead = {tag_field: "redhead", "name": "Jack", "age": 50}

    assert serialize(U, blonde, Context()) == data_blonde
    assert serialize(U, redhead, Context()) == data_redhead

    assert deserialize(U, data_blonde) == blonde
    assert deserialize(U, data_redhead) == redhead


def test_tag_validationerrors():
    lt = Tag("foo", Blonde)

    with pytest.raises(ValidationError, match="Tag 'foo' is required"):
        lt.get_type(None, ctx=None)

    with pytest.raises(ValidationError, match="Tag 'bar' does not match expected tag 'foo'"):
        lt.get_type("bar", ctx=None)

    with pytest.raises(
        ValidationError, match="Type '.*' does not match expected class '.*Blonde.*'"
    ):
        lt.get_tag(Redhead, ctx=None)



from dataclasses import dataclass

import pytest

from serieux import Serieux
from serieux.exc import ValidationError
from serieux.features.lazy import DeepLazy, Lazy, LazyDeserialization, LazyProxy
from serieux.features.partial import Sources

from .definitions import Point

deserialize = (Serieux + LazyDeserialization)().deserialize


@dataclass(frozen=True)
class Person:
    name: str
    age: int

    def __post_init__(self):
        if self.age < 0:
            raise ValueError("Age cannot be negative!")


def test_lazy_object():
    person = Person(name="Bob", age=78)
    lazy_person = LazyProxy(lambda: person)

    assert person == lazy_person
    assert lazy_person.name == "Bob"
    assert lazy_person.age == 78
    assert hash(person) == hash(lazy_person)


def test_laziness():
    data = {"name": "Clara", "age": -10}
    lazy = deserialize(Lazy[Person], data)
    with pytest.raises(ValidationError, match="Age cannot be negative"):
        lazy.age


def test_deep_laziness():
    data = [
        {"name": "Alice", "age": 18},
        {"name": "Bob", "age": -78},
        {"name": "Clara", "age": 10},
    ]
    lazy = deserialize(Lazy[list[Person]], data)
    with pytest.raises(ValidationError, match="Age cannot be negative"):
        lazy[0].name

    deep_lazy = deserialize(DeepLazy[list[Person]], data)
    assert deep_lazy[0].name == "Alice"
    deep_lazy[1]  # No attributes fetched, so nothing happens
    with pytest.raises(ValidationError, match="Age cannot be negative"):
        deep_lazy[1].name
    assert deep_lazy[2].name == "Clara"


@dataclass
class ContainsLazy:
    normal: str
    pt: Lazy[Point]


def test_lazy_partial_invalid():
    result = deserialize(ContainsLazy, Sources({"normal": "hello", "pt": {"x": 1}}))
    assert result.normal == "hello"
    with pytest.raises(ValidationError):
        result.pt.x


def test_lazy_partial():
    result = deserialize(
        ContainsLazy,
        Sources(
            {"normal": "hello", "pt": {"x": 1}},
            {"pt": {"y": 18}},
        ),
    )
    assert result.normal == "hello"
    assert result.pt.x == 1
    assert result.pt.y == 18
    assert type(result.pt.y) is int


def test_lazy_callable():
    def add(a, b):
        return a + b

    lazy_func = LazyProxy(lambda: add)

    assert lazy_func(1, 2) == 3
    assert lazy_func(a=1, b=2) == 3

    # Test with a class that implements __call__
    class Adder:
        def __init__(self, base):
            self.base = base

        def __call__(self, x):
            return self.base + x

    lazy_adder = LazyProxy(lambda: Adder(10))
    assert lazy_adder(5) == 15



import marshmallow_dataclass
import orjson as json
from apischema import deserialize as apischema_deserialize, serialize as apischema_serialize
from mashumaro.codecs.basic import BasicDecoder, BasicEncoder
from mashumaro.codecs.orjson import ORJSONEncoder
from pydantic import TypeAdapter
from serde.de import from_dict as serde_from_dict
from serde.json import to_json as serde_to_json
from serde.se import to_dict as serde_to_dict

from serieux import deserialize as serieux_deserialize, serialize as serieux_serialize, serieux
from serieux.ctx import EmptyContext, empty


class SerieuxInterface:
    __name__ = "serieux"

    def serializer_for_type(self, t):
        func = serieux_serialize.resolve(type[t], t, EmptyContext)
        return lambda x: func(serieux, t, x, empty)

    def json_for_type(self, t):
        func = serieux_serialize.resolve(type[t], t, EmptyContext)
        return lambda x: json.dumps(func(serieux, t, x, empty))

    def deserializer_for_type(self, t):
        func = serieux_deserialize.resolve(type[t], dict, EmptyContext)
        return lambda x: func(serieux, t, x, empty)


serieux = SerieuxInterface()


class ApischemaInterface:
    __name__ = "apischema"

    def serializer_for_type(self, t):
        return lambda x: apischema_serialize(t, x, check_type=False)

    def json_for_type(self, t):
        return lambda x: json.dumps(apischema_serialize(t, x, check_type=False))

    def deserializer_for_type(self, t):
        return lambda x: apischema_deserialize(t, x)


apischema = ApischemaInterface()


class PydanticInterface:
    __name__ = "pydantic"

    def serializer_for_type(self, t):
        return TypeAdapter(t).serializer.to_python

    def json_for_type(self, t):
        return TypeAdapter(t).serializer.to_json

    def deserializer_for_type(self, t):
        return TypeAdapter(t).validator.validate_python


pydantic = PydanticInterface()


class MarshmallowInterface:
    __name__ = "marshmallow"

    def serializer_for_type(self, t):
        schema = marshmallow_dataclass.class_schema(t)()
        return lambda x: schema.dump(x)

    def json_for_type(self, t):
        schema = marshmallow_dataclass.class_schema(t)()
        return lambda x: json.dumps(schema.dump(x))

    def deserializer_for_type(self, t):
        schema = marshmallow_dataclass.class_schema(t)()
        return lambda x: schema.load(x)


marshmallow = MarshmallowInterface()


class MashumaroInterface:
    __name__ = "mashumaro"

    def serializer_for_type(self, t):
        return BasicEncoder(t).encode

    def json_for_type(self, t):
        return ORJSONEncoder(t).encode

    def deserializer_for_type(self, t):
        return BasicDecoder(t).decode


mashumaro = MashumaroInterface()


class SerdeInterface:
    __name__ = "serde"

    def serializer_for_type(self, t):
        return serde_to_dict

    def json_for_type(self, t):
        return serde_to_json

    def deserializer_for_type(self, t):
        return lambda x: serde_from_dict(t, x)


serde = SerdeInterface()



import pytest
from ovld import Medley

from serieux import deserializer, schema_definition, serializer
from serieux.ctx import Context
from serieux.exc import ValidationError


class Beep:
    def __init__(self, value: int):
        self.value = value

    def __eq__(self, other):
        return self.value == other.value


def test_custom_serializer(fresh_serieux):
    with pytest.raises(ValidationError):
        fresh_serieux.serialize(Beep, Beep(7))

    @serializer
    def _s(self, t: type[Beep], obj: Beep, ctx: Context):
        return obj.value

    @serializer(priority=-1)
    def _s(self, t: type[Beep], obj: Beep, ctx: Context):
        assert False, "should not use this method"

    assert fresh_serieux.serialize(Beep, Beep(7)) == 7


def test_custom_deserializer(fresh_serieux):
    with pytest.raises(ValidationError):
        fresh_serieux.deserialize(Beep, 7)

    @deserializer
    def _d(self, t: type[Beep], obj: int, ctx: Context):
        return Beep(obj)

    @deserializer(priority=-1)
    def _d(self, t: type[Beep], obj: int, ctx: Context):
        assert False, "should not use this method"

    assert fresh_serieux.deserialize(list[Beep], [1, 2, 3]) == [Beep(1), Beep(2), Beep(3)]


def test_custom_schema(fresh_serieux):
    with pytest.raises(Exception):
        fresh_serieux.schema(Beep)

    @schema_definition
    def _sch(self, t: type[Beep], ctx: Context):
        return {"type": "integer"}

    @schema_definition(priority=-1)
    def _sch(self, t: type[Beep], ctx: Context):
        assert False, "should not use this method"

    assert fresh_serieux.schema(Beep).compile(root=False) == {"type": "integer"}


class Kustom(Medley):
    def deserialize(self, t: type[Beep], obj: str, ctx: Context):
        return Beep(int(obj))


def test_inherits(fresh_serieux):
    srx = fresh_serieux + Kustom()

    with pytest.raises(ValidationError):
        srx.deserialize(Beep, 7)
    assert srx.deserialize(Beep, "9") == Beep(9)

    @deserializer
    def _d(self, t: type[Beep], obj: int, ctx: Context):
        return Beep(obj)

    assert srx.deserialize(Beep, 7) == Beep(7)
    assert srx.deserialize(Beep, "9") == Beep(9)



import ast
import inspect
import sys
import traceback
from ast import NodeTransformer
from contextlib import contextmanager
from textwrap import dedent

import pytest
from _pytest.assertion.rewrite import AssertionRewriter

from serieux.exc import ValidationExceptionGroup


@contextmanager
def validation_errors(msgs):
    try:
        yield
    except ValidationExceptionGroup as veg:
        for pth, msg in msgs.items():
            if not any((exc.access_string() == pth and msg in str(exc)) for exc in veg.exceptions):
                traceback.print_exception(veg)
                raise Exception(f"No exception was raised at {pth} for '{msg}'")


class AssertTransformer(NodeTransformer):
    def visit_FunctionDef(self, node):
        newfns = []
        for i, stmt in enumerate(node.body):
            if not isinstance(stmt, ast.Assert):
                raise Exception("@one_test_per_assert requires all statements to be asserts")
            else:
                newfns.append(
                    ast.FunctionDef(
                        name=f"{node.name}_assert{i + 1}",
                        args=node.args,
                        body=[stmt],
                        decorator_list=node.decorator_list,
                        returns=node.returns,
                    )
                )
        return ast.Module(body=newfns, type_ignores=[])


def one_test_per_assert(fn):
    src = dedent(inspect.getsource(fn))
    filename = inspect.getsourcefile(fn)
    tree = ast.parse(src, filename)
    tree = tree.body[0]
    assert isinstance(tree, ast.FunctionDef)
    tree.decorator_list = []
    new_tree = AssertTransformer().visit(tree)
    ast.fix_missing_locations(new_tree)
    _, lineno = inspect.getsourcelines(fn)
    ast.increment_lineno(new_tree, lineno - 1)
    # Use pytest's assertion rewriter for nicer error messages
    AssertionRewriter(filename, None, None).run(new_tree)
    new_fn = compile(new_tree, filename, "exec")
    glb = fn.__globals__
    exec(new_fn, glb, glb)
    return None


has_312_features = pytest.mark.skipif(
    sys.version_info < (3, 12), reason="This test relies on Python 3.12+ syntax"
)



import hashlib
import json
import tomllib
import uuid
from pathlib import Path
from typing import Any

import yaml
from ovld import call_next, dependent_check, ovld, recurse
from ovld.dependent import HasKey

from ..ctx import Context, Located, Location
from ..exc import ValidationError
from ..priority import HI7, MIN
from ..utils import clsstring
from .partial import PartialBuilding, Sources

include_field = "$include"


@dependent_check
def FileSuffix(value: Path, *suffixes):
    return value.exists() and value.suffix in suffixes


@ovld
def parse(path: FileSuffix[".toml"]):  # type: ignore
    return tomllib.loads(path.read_text())


@ovld
def parse(path: FileSuffix[".json"]):  # type: ignore
    return json.loads(path.read_text())


@ovld
def parse(path: FileSuffix[".yaml", ".yml"]):  # type: ignore
    return yaml.compose(path.read_text()) or {}


@ovld
def parse(path: Path):
    raise ValidationError(f"Could not read data from file '{path}'")


class WorkingDirectory(Context):
    directory: Path = None
    origin: Path = None

    def __post_init__(self):
        if self.directory is None:
            self.directory = self.origin.parent

    def make_path_for(self, *, name=None, suffix=None, entropy=None):
        if name is None and entropy is not None:
            name = hashlib.md5(
                str(entropy).encode() if isinstance(entropy, str) else entropy
            ).hexdigest()
        if name is None:
            name = str(uuid.uuid4())
        pth = self.directory / name
        if suffix is not None:
            pth = pth.with_suffix(suffix)
        return pth

    def save_to_file(
        self, data: str | bytes = None, suffix=None, *, name=None, callback=None, entropy=None
    ):
        dest = self.make_path_for(entropy=entropy or data, suffix=suffix, name=name)
        dest.parent.mkdir(parents=True, exist_ok=True)
        if callback:
            callback(dest)
        else:
            if isinstance(data, str):
                mode = "w"
                encoding = "utf-8"
            else:
                mode = "wb"
                encoding = None
            with open(dest, mode=mode, encoding=encoding) as f:
                f.write(data)
        return str(dest.relative_to(self.directory))


def yaml_source_extract(node, ctx):
    return ctx + Located(
        location=Location(
            source=getattr(ctx, "origin", None),
            code=node.start_mark.buffer,
            start=node.start_mark.index,
            end=node.end_mark.index,
            linecols=(
                (node.start_mark.line, node.start_mark.column),
                (node.end_mark.line, node.end_mark.column),
            ),
        )
    )


@dependent_check
def ScalarNode(value: yaml.ScalarNode, tag_suffix):
    return value.tag.endswith(tag_suffix)


class FromFile(PartialBuilding):
    def deserialize(self, t: Any, obj: Path, ctx: Context):
        if isinstance(ctx, WorkingDirectory):
            obj = ctx.directory / obj.expanduser()
        data = parse(obj)
        ctx = ctx + WorkingDirectory(origin=obj, directory=obj.parent)
        return recurse(t, data, ctx)

    @ovld(priority=HI7)
    def deserialize(self, t: Any, obj: yaml.MappingNode, ctx: Context):
        return recurse(t, {k.value: v for k, v in obj.value}, yaml_source_extract(obj, ctx))

    @ovld(priority=HI7)
    def deserialize(self, t: Any, obj: yaml.SequenceNode, ctx: Context):
        return recurse(t, obj.value, yaml_source_extract(obj, ctx))

    @ovld(priority=HI7)
    def deserialize(self, t: Any, obj: ScalarNode[":str"], ctx: Context):  # type: ignore
        return recurse(t, obj.value, yaml_source_extract(obj, ctx))

    @ovld(priority=HI7)
    def deserialize(self, t: Any, obj: ScalarNode[":int"], ctx: Context):  # type: ignore
        return recurse(t, int(obj.value), yaml_source_extract(obj, ctx))

    @ovld(priority=HI7)
    def deserialize(self, t: Any, obj: ScalarNode[":float"], ctx: Context):  # type: ignore
        return recurse(t, float(obj.value), yaml_source_extract(obj, ctx))

    @ovld(priority=HI7)
    def deserialize(self, t: Any, obj: ScalarNode[":bool"], ctx: Context):  # type: ignore
        value = obj.value.lower() in ("yes", "on", "true")
        return recurse(t, value, yaml_source_extract(obj, ctx))

    @ovld(priority=HI7)
    def deserialize(self, t: Any, obj: ScalarNode[":null"], ctx: Context):  # type: ignore
        return recurse(t, None, yaml_source_extract(obj, ctx))

    @ovld(priority=HI7)
    def deserialize(self, t: Any, obj: ScalarNode[":timestamp"], ctx: Context):  # type: ignore
        return recurse(t, obj.value, yaml_source_extract(obj, ctx))

    @ovld(priority=HI7)
    def deserialize(self, t: Any, obj: yaml.ScalarNode, ctx: Context):  # pragma: no cover
        raise ValidationError(f"Cannot deserialize YAML node of type `{obj.tag}`")


class IncludeFile(FromFile):
    @ovld(priority=1)
    def deserialize(self, t: type[object], obj: HasKey[include_field], ctx: Context):
        obj = dict(obj)
        incl = recurse(str, obj.pop(include_field), ctx)
        return recurse(t, Sources(Path(incl), obj), ctx)

    @ovld(priority=MIN)
    def deserialize(self, t: type[object], obj: str, ctx: WorkingDirectory):
        if "." not in obj or obj.rsplit(".", 1)[1].isnumeric():
            return call_next(t, obj, ctx)

        path = ctx.directory / obj
        if path.exists():
            return recurse(t, path, ctx)
        else:
            raise ValidationError(
                f"Tried to read '{obj}' as a configuration file (at path '{path}')"
                f" to deserialize into object of type {clsstring(t)},"
                " but there was no such file."
            )



from dataclasses import dataclass
from pathlib import Path
from types import NoneType
from typing import Annotated

from ovld import Code, ovld, recurse

from .instructions import pushdown
from .model import FieldModelizable, StringModelizable, model
from .utils import TypeAliasType, basic_type


class Tell:
    def __lt__(self, other):
        return self.cost() < other.cost()

    def cost(self):
        return 1


@dataclass(frozen=True)
class TypeTell(Tell):
    t: type

    def gen(self, arg):
        return Code("isinstance($arg, $t)", arg=arg, t=self.t)


@dataclass(frozen=True)
class KeyTell(Tell):
    key: str

    def gen(self, arg):
        return Code("(isinstance($arg, dict) and $k in $arg)", arg=arg, k=self.key)

    def cost(self):
        return 2


@dataclass(frozen=True)
class KeyValueTell(Tell):
    key: str
    value: object

    def gen(self, arg):
        return Code(
            "(isinstance($arg, dict) and $k in $arg and $arg[$k] == $v)",
            arg=arg,
            k=self.key,
            v=self.value,
        )

    def cost(self):  # pragma: no cover
        return 3


@ovld
def tells(
    typ: type[int]
    | type[str]
    | type[bool]
    | type[float]
    | type[NoneType]
    | type[list]
    | type[dict],
):
    return {TypeTell(basic_type(typ))}


@ovld
def tells(typ: type[Path]):
    return {TypeTell(str)}


@ovld
def tells(typ: type[FieldModelizable]):
    m = model(typ)
    return {TypeTell(dict)} | {KeyTell(f.serialized_name) for f in m.fields}


@ovld
def tells(typ: type[StringModelizable]):
    return {TypeTell(str)}


@ovld
def tells(typ: TypeAliasType):
    return recurse(typ.__value__)


@ovld(priority=-1)
def tells(typ: type[Annotated]):
    return recurse(pushdown(typ))



from dataclasses import MISSING, field
from datetime import date, datetime
from enum import Enum
from functools import wraps
from itertools import pairwise
from pathlib import Path
from types import NoneType, UnionType
from typing import Annotated, Any, get_args, get_origin

import yaml
from ovld import (
    Code,
    CodegenInProgress,
    CodegenParameter,
    Def,
    Lambda,
    Medley,
    call_next,
    code_generator,
    keyword_decorator,
    ovld,
    recurse,
)
from ovld.codegen import Function
from ovld.medley import KeepLast, use_combiner
from ovld.types import All
from ovld.utils import ResolutionError, subtler_type

from .ctx import AccessPath, Context
from .exc import SchemaError, SerieuxError, ValidationError, ValidationExceptionGroup
from .features.fromfile import WorkingDirectory
from .instructions import pushdown
from .model import FieldModelizable, Modelizable, StringModelizable, model
from .priority import LOW, MAX, MIN, STD, STD2, STD3
from .schema import AnnotatedSchema, Schema
from .tell import tells as get_tells
from .utils import (
    Indirect,
    IsLiteral,
    TypeAliasType,
    UnionAlias,
    basic_type,
    clsstring,
)


@keyword_decorator
def code_generator_wrap_error(fn, priority=0):
    @wraps(fn)
    def f(cls, *args, **kwargs):
        result = fn(cls, *args, **kwargs)
        if not result:
            return None
        body = result.create_body(("t", "obj", "ctx"))
        stmts = [
            "try:",
            [body],
            "except $SXE:",
            ["raise"],
            "except Exception as exc:",
            ["raise $VE(exc=exc, ctx=$ctx)"],
        ]
        return Def(stmts, SXE=SerieuxError, VE=ValidationError)

    return code_generator(f, priority=priority)


class BaseImplementation(Medley):
    default_context: Context = field(default_factory=AccessPath)
    validate_serialize: CodegenParameter[bool] = True
    validate_deserialize: CodegenParameter[bool] = True

    def __post_init__(self):
        self._schema_cache = {}

    #######################
    # User-facing methods #
    #######################

    @use_combiner(KeepLast)
    def load(self, t, obj, ctx=None):
        ctx = ctx or self.default_context
        return self.deserialize(t, obj, ctx)

    @use_combiner(KeepLast)
    def dump(self, t, obj, ctx=None, *, dest=None):
        ctx = ctx or self.default_context
        if dest:
            dest = Path(dest)
            ctx = ctx + WorkingDirectory(origin=dest)
        serialized = self.serialize(t, obj, ctx)
        if dest:
            assert dest.suffix == ".yaml"  # TODO: support other formats
            dest.parent.mkdir(parents=True, exist_ok=True)
            dest.write_text(yaml.safe_dump(serialized, sort_keys=False))
        else:
            return serialized

    ##################
    # Global helpers #
    ##################

    @classmethod
    def subcode(
        cls, method_name, t, accessor, ctx_t, ctx_expr=Code("$ctx"), after=None, validate=None
    ):
        accessor = accessor if isinstance(accessor, Code) else Code(accessor)
        if validate is None:
            validate = getattr(cls, f"validate_{method_name}")
        method = getattr(cls, method_name)
        if ec := getattr(cls, f"{method_name}_embed_condition")(t):
            ec = All[ec]
        if ec is not None:
            try:
                fn = method.resolve(type[t], ec, ctx_t, after=after)
                cg = getattr(fn, "__codegen__", None)
                if cg:
                    body = cg.create_expression([None, t, accessor, ctx_expr])
                    ot = get_origin(t) or t
                    if not validate:
                        return Code(body)
                    else:
                        return Code(
                            "$body if type(__checked := $accessor) is $t else $recurse($self, $t, __checked, $ctx_expr)",
                            body=Code(body),
                            accessor=accessor,
                            t=ot,
                            recurse=method,
                            ctx_expr=ctx_expr,
                        )
            except (CodegenInProgress, ValueError):  # pragma: no cover
                # This is important if we are going to inline recursively
                # a type that refers to itself down the line.
                # We currently never do that.
                pass
        return Code(
            "$method_map[$tt, type(OBJ := $accessor), $ctxt]($self, $t, OBJ, $ctx_expr)",
            tt=subtler_type(t),
            ctxt=ctx_t,
            t=t,
            accessor=accessor,
            method_map=method.map,
            ctx_expr=ctx_expr,
        )

    ########################################
    # serialize:  helpers and entry points #
    ########################################

    @classmethod
    def serialize_embed_condition(cls, t):
        if t in (int, str, bool, float, NoneType):
            return t

    @ovld(priority=MIN)
    def serialize(self, t: Any, obj: Any, ctx: Context, /):
        raise ValidationError(
            f"Cannot serialize object of type '{clsstring(type(obj))}'"
            + (f" into expected type '{clsstring(t)}'." if t is not type(obj) else ""),
            ctx=ctx,
        )

    @ovld(priority=MIN)
    def serialize(self, t: Indirect | TypeAliasType, obj: Any, ctx: Context, /):
        return recurse(t.__value__, obj, ctx)

    @ovld(priority=LOW)
    def serialize(self, t: type[Annotated], obj: Any, ctx: Context, /):
        return recurse(pushdown(t), obj, ctx)

    def serialize(self, obj: Any, /):
        return recurse(type(obj), obj, self.default_context)

    def serialize(self, t: Any, obj: Any, /):
        return recurse(t, obj, self.default_context)

    #########################################
    # deserialize: helpers and entry points #
    #########################################

    deserialize_embed_condition = serialize_embed_condition

    @ovld(priority=MIN)
    def deserialize(self, t: Any, obj: Any, ctx: Context, /):
        try:
            # Pass through if the object happens to already be the right type
            if t is Any or isinstance(obj, t):
                return obj
        except TypeError:  # pragma: no cover
            pass
        disp = str(obj)
        if len(disp) > (n := 30):  # pragma: no cover
            disp = disp[:n] + "..."
        if isinstance(obj, str):
            descr = f"string '{disp}'"
        else:
            descr = f"object `{disp}` of type '{clsstring(type(obj))}'"
        raise ValidationError(
            f"Cannot deserialize {descr} into expected type '{clsstring(t)}'.",
            ctx=ctx,
        )

    @ovld(priority=MIN)
    def deserialize(self, t: Indirect | TypeAliasType, obj: Any, ctx: Context, /):
        return recurse(t.__value__, obj, ctx)

    @ovld(priority=LOW)
    def deserialize(self, t: type[Annotated], obj: Any, ctx: Context, /):
        return recurse(pushdown(t), obj, ctx)

    def deserialize(self, t: Any, obj: Any, /):
        return recurse(t, obj, self.default_context)

    ####################################
    # schema: helpers and entry points #
    ####################################

    @ovld(priority=MAX)
    def schema(self, t: Any, ctx: Context, /):
        if t not in self._schema_cache:
            self._schema_cache[t] = holder = Schema(t)
            try:
                result = call_next(t, ctx)
            except Exception:
                del self._schema_cache[t]
                raise
            holder.update(result)
        return self._schema_cache[t]

    @ovld(priority=MIN)
    def schema(self, t: Indirect | TypeAliasType, ctx: Context, /):
        return recurse(t.__value__, ctx)

    @ovld(priority=LOW)
    def schema(self, t: type[Annotated], ctx: Context, /):
        return recurse(pushdown(t), ctx)

    def schema(self, t: Any, /):
        return recurse(t, self.default_context)

    ################################
    # Implementations: basic types #
    ################################

    # str

    @code_generator(priority=STD)
    def serialize(cls, t: type[str], obj: str, ctx: Context, /):
        return Lambda(Code("$obj"))

    @code_generator(priority=STD)
    def deserialize(cls, t: type[str], obj: str, ctx: Context, /):
        return Lambda(Code("$obj"))

    @ovld(priority=STD)
    def schema(self, t: type[str], ctx: Context, /):
        return {"type": "string"}

    # bool

    @code_generator(priority=STD)
    def serialize(cls, t: type[bool], obj: bool, ctx: Context, /):
        return Lambda(Code("$obj"))

    @code_generator(priority=STD)
    def deserialize(cls, t: type[bool], obj: bool, ctx: Context, /):
        return Lambda(Code("$obj"))

    @ovld(priority=STD)
    def schema(self, t: type[bool], ctx: Context, /):
        return {"type": "boolean"}

    # int

    @code_generator(priority=STD)
    def serialize(cls, t: type[int], obj: int, ctx: Context, /):
        return Lambda(Code("$obj"))

    @code_generator(priority=STD)
    def deserialize(cls, t: type[int], obj: int, ctx: Context, /):
        return Lambda(Code("$obj"))

    @ovld(priority=STD)
    def schema(self, t: type[int], ctx: Context, /):
        return {"type": "integer"}

    # float

    @code_generator(priority=STD)
    def serialize(cls, t: type[float], obj: float, ctx: Context, /):
        return Lambda(Code("$obj"))

    @code_generator(priority=STD)
    def deserialize(cls, t: type[float], obj: float, ctx: Context, /):
        return Lambda(Code("$obj"))

    # float

    @code_generator(priority=STD)
    def serialize(cls, t: type[float], obj: int, ctx: Context, /):
        return Lambda(Code("$obj"))

    @code_generator(priority=STD)
    def deserialize(cls, t: type[float], obj: int, ctx: Context, /):
        return Lambda(Code("float($obj)"))

    @ovld(priority=STD)
    def schema(self, t: type[float], ctx: Context, /):
        return {"type": "number"}

    # None

    @code_generator(priority=STD)
    def serialize(cls, t: type[NoneType], obj: NoneType, ctx: Context, /):
        return Lambda(Code("$obj"))

    @code_generator(priority=STD)
    def deserialize(cls, t: type[NoneType], obj: NoneType, ctx: Context, /):
        return Lambda(Code("$obj"))

    @ovld(priority=STD)
    def schema(self, t: type[NoneType], ctx: Context, /):
        return {"type": "null"}

    ##########################
    # Implementations: lists #
    ##########################

    @classmethod
    def __generic_codegen_list(cls, method, t, obj, ctx):
        (t,) = get_args(t)
        builder = list if method == "serialize" else get_origin(t) or t
        (lt,) = get_args(t) or (object,)
        comp = "$lbody for IDX, X in enumerate($obj)"
        if builder is list:
            code = f"[{comp}]"
        elif builder is set:
            code = f"{{{comp}}}"
        else:
            code = f"$builder({comp})"
        if hasattr(ctx, "follow"):
            ctx_expr = Code("$ctx.follow($objt, $obj, IDX)", objt=t)
            return Lambda(
                code,
                lbody=cls.subcode(method, lt, "X", ctx, ctx_expr=ctx_expr),
                builder=builder,
            )
        else:
            return Lambda("[$lbody for X in $obj]", lbody=cls.subcode(method, lt, "X", ctx))

    @code_generator(priority=STD)
    def serialize(cls, t: type[list], obj: list, ctx: Context, /):
        return cls.__generic_codegen_list("serialize", t, obj, ctx)

    @code_generator(priority=STD)
    def deserialize(cls, t: type[list], obj: list, ctx: Context, /):
        return cls.__generic_codegen_list("deserialize", t, obj, ctx)

    def schema(self, t: type[list], ctx: Context, /):
        (lt,) = get_args(t)
        follow = hasattr(ctx, "follow")
        fctx = ctx.follow(t, None, "*") if follow else ctx
        return {"type": "array", "items": recurse(lt, fctx)}

    #########################
    # Implementations: sets #
    #########################

    @code_generator(priority=STD)
    def serialize(cls, t: type[set], obj: set, ctx: Context, /):
        return cls.__generic_codegen_list("serialize", t, obj, ctx)

    @code_generator(priority=STD)
    def deserialize(cls, t: type[set], obj: list, ctx: Context, /):
        return cls.__generic_codegen_list("deserialize", t, obj, ctx)

    ###############################
    # Implementations: frozensets #
    ###############################

    @code_generator(priority=STD)
    def serialize(cls, t: type[frozenset], obj: frozenset, ctx: Context, /):
        return cls.__generic_codegen_list("serialize", t, obj, ctx)

    @code_generator(priority=STD)
    def deserialize(cls, t: type[frozenset], obj: list, ctx: Context, /):
        return cls.__generic_codegen_list("deserialize", t, obj, ctx)

    ##########################
    # Implementations: dicts #
    ##########################

    @classmethod
    def __generic_codegen_dict(cls, method, t: type[dict], obj: dict, ctx: Context, /):
        (t,) = get_args(t)
        builder = dict if method == "serialize" else get_origin(t) or t
        kt, vt = get_args(t) or (object, object)
        ctx_expr = (
            Code("$ctx.follow($objt, $obj, K)", objt=t) if hasattr(ctx, "follow") else Code("$ctx")
        )
        code = "{$kbody: $vbody for K, V in $obj.items()}"
        if builder is not dict:
            code = f"$builder({code})"
        return Lambda(
            code,
            kbody=cls.subcode(method, kt, "K", ctx, ctx_expr=ctx_expr),
            vbody=cls.subcode(method, vt, "V", ctx, ctx_expr=ctx_expr),
            builder=builder,
        )

    @code_generator(priority=STD)
    def serialize(cls, t: type[dict], obj: dict, ctx: Context, /):
        return cls.__generic_codegen_dict("serialize", t, obj, ctx)

    @code_generator(priority=STD)
    def deserialize(cls, t: type[dict], obj: dict, ctx: Context, /):
        return cls.__generic_codegen_dict("deserialize", t, obj, ctx)

    def schema(self, t: type[dict], ctx: Context, /):
        kt, vt = get_args(t)
        if kt is not str:
            raise SchemaError(
                f"Cannot create a schema for dicts with non-string keys (found key type: `{kt}`)"
            )
        follow = hasattr(ctx, "follow")
        fctx = ctx.follow(t, None, "*") if follow else ctx
        return {"type": "object", "additionalProperties": recurse(vt, fctx)}

    #####################################
    # Implementations: FieldModelizable #
    #####################################

    @code_generator_wrap_error(priority=STD)
    def serialize(cls, t: type[FieldModelizable], obj: Any, ctx: Context, /):
        (orig_t,) = get_args(t)
        t = model(orig_t)
        if not t.accepts(obj):
            return None
        stmts = []
        follow = hasattr(ctx, "follow")
        for i, f in enumerate(t.fields):
            if f.property_name is None:
                raise SchemaError(
                    f"Cannot serialize '{clsstring(t)}' because its model does not specify how to serialize property '{f.name}'"
                )
            ctx_expr = (
                Code("$ctx.follow($objt, $obj, $fld)", objt=orig_t, fld=f.name)
                if follow
                else Code("$ctx")
            )
            stmt = Code(
                f"v_{i} = $setter",
                setter=cls.subcode(
                    "serialize", f.type, f"$obj.{f.property_name}", ctx, ctx_expr=ctx_expr
                ),
            )
            stmts.append(stmt)
        final = Code(
            "return {$[,]parts}",
            parts=[
                Code(
                    f"$fname: v_{i}",
                    fname=f.serialized_name,
                )
                for i, f in enumerate(t.fields)
            ],
        )
        stmts.append(final)
        return Def(stmts, VE=ValidationError, VEG=ValidationExceptionGroup)

    @code_generator_wrap_error(priority=STD)
    def deserialize(cls, t: type[FieldModelizable], obj: dict, ctx: Context, /):
        (orig_t,) = get_args(t)
        t = model(orig_t)
        follow = hasattr(ctx, "follow")
        stmts = []
        args = []
        for i, f in enumerate(t.fields):
            ctx_expr = (
                Code("$ctx.follow($objt, $obj, $fld)", objt=orig_t, fld=f.name)
                if follow
                else Code("$ctx")
            )
            if f.metavar:
                expr = Code(f.metavar)
            else:
                expr = cls.subcode(
                    "deserialize",
                    f.type,
                    Code("$obj[$pname]", pname=f.serialized_name),
                    ctx,
                    ctx_expr=ctx_expr,
                )

            stmt = Code([f"v_{i} = $expr"], expr=expr)

            if f.required or f.metavar:
                pass
            elif f.default is not MISSING:
                stmt = Code(
                    [
                        "if $pname in $obj:",
                        ["$stmt"],
                        "else:",
                        [f"v_{i} = $dflt"],
                    ],
                    dflt=f.default,
                    pname=f.serialized_name,
                    stmt=stmt,
                )
            elif f.default_factory is not MISSING:
                stmt = Code(
                    [
                        "if $pname in $obj:",
                        ["$stmt"],
                        "else:",
                        [f"v_{i} = $dflt()"],
                    ],
                    dflt=f.default_factory,
                    pname=f.serialized_name,
                    stmt=stmt,
                )
            stmts.append(stmt)
            if isinstance(f.argument_name, str):
                arg = f"{f.argument_name}=v_{i}"
            else:
                arg = f"v_{i}"
            args.append(arg)

        final = Code(
            "return $constructor($[,]parts)",
            constructor=t.constructor,
            parts=[Code(a) for a in args],
        )
        stmts.append(final)
        return Def(stmts, VE=ValidationError, VEG=ValidationExceptionGroup)

    ######################################
    # Implementations: StringModelizable #
    ######################################

    @code_generator_wrap_error(priority=STD2)
    def serialize(self, t: type[StringModelizable], obj: Any, ctx: Context, /):
        (t,) = get_args(t)
        m = model(t)
        if not m.accepts(obj) or m.to_string is None:
            return None
        if isinstance(m.to_string, Function):
            return m.to_string
        else:
            return Lambda("$to_string($obj)", to_string=m.to_string)

    @code_generator_wrap_error(priority=STD2)
    def deserialize(self, t: type[StringModelizable], obj: str, ctx: Context, /):
        (t,) = get_args(t)
        m = model(t)
        if m.regexp:
            if isinstance(m.from_string, Def):  # pragma: no cover
                raise Exception("In model definitions, use Lambda with regexp, not Def")
            elif isinstance(m.from_string, Lambda):
                expr = m.from_string.create_expression(["t", "obj", "ctx"])
            else:
                expr = Code("$from_string($obj)", from_string=m.from_string)
            descr = m.string_description or f"pattern {m.regexp.pattern!r}"
            pattern = f"String {{$obj!r}} is not a valid {clsstring(t)}. It should match: {descr}"
            return Def(
                ["t", "obj", "ctx"],
                Code(
                    [
                        ["if $regexp.match($obj):", ["return $expr"]],
                        [
                            "else:",
                            [f"""raise $VE(f"{pattern}")"""],
                        ],
                    ]
                ),
                from_string=m.from_string,
                regexp=m.regexp,
                expr=expr,
                descr=descr,
                VE=ValidationError,
            )
        elif isinstance(m.from_string, Function):
            return m.from_string
        else:
            return Lambda("$from_string($obj)", from_string=m.from_string)

    ################################
    # Implementations: Modelizable #
    ################################

    @ovld(priority=STD)
    def schema(self, t: type[Modelizable], ctx: Context, /):
        m = model(t)

        f_schema, s_schema = None, None
        follow = hasattr(ctx, "follow")

        if m.fields is not None:
            properties = {}
            required = []

            for f in m.fields:
                fctx = ctx.follow(t, None, f.name) if follow else ctx
                fsch = recurse(f.type, fctx)
                extra = {}
                if f.description:
                    extra["description"] = f.description
                if f.default is not MISSING:
                    extra["default"] = f.default
                fsch = fsch if not f.description else AnnotatedSchema(fsch, **extra)
                properties[f.serialized_name] = fsch
                if f.required:
                    required.append(f.serialized_name)

            f_schema = {
                "type": "object",
                "properties": properties,
                "required": required,
                "additionalProperties": m.extensible,
            }

        if m.from_string is not None:
            s_schema = {"type": "string"}
            if m.regexp:
                s_schema["pattern"] = m.regexp.pattern

        assert f_schema or s_schema

        if f_schema is not None and s_schema is not None:
            return {"oneOf": [f_schema, s_schema]}
        elif f_schema is not None:
            return f_schema
        else:
            return s_schema

    ###########################
    # Implementations: Unions #
    ###########################

    @code_generator(priority=STD)
    def serialize(cls, t: type[UnionAlias], obj: Any, ctx: Context, /):
        (t,) = get_args(t)
        o1, *rest = get_args(t)
        code = cls.subcode("serialize", o1, "$obj", ctx, validate=False)
        for opt in rest:
            code = Code(
                "$ocode if isinstance($obj, $sopt) else $code",
                sopt=basic_type(opt),
                ocode=cls.subcode("serialize", opt, "$obj", ctx, validate=False),
                code=code,
            )
        return Lambda(code)

    @code_generator(priority=STD)
    def deserialize(cls, t: type[UnionAlias] | type[UnionType], obj: Any, ctx: Context, /):
        (t,) = get_args(t)
        options = get_args(t)

        try:
            tells = [get_tells(opt := o) for o in options]
        except ResolutionError as exc:
            raise SchemaError(
                f"Cannot deserialize union type `{t}`, because no rule is defined to discriminate `{opt}` from other types.",
                exc=exc,
            )
        elim = set()
        for tl1, tl2 in pairwise(tells):
            elim |= tl1 & tl2
        for tls in tells:
            tls -= elim

        if sum(not tl for tl in tells) > 1:
            raise SchemaError(f"Cannot differentiate the possible union members in type '{t}'")

        options = list(zip(tells, options))
        options.sort(key=lambda x: len(x[0]))

        (_, o1), *rest = options

        code = cls.subcode("deserialize", o1, "$obj", ctx)
        for tls, opt in rest:
            code = Code(
                "($ocode if $cond else $code)",
                cond=min(tls).gen(Code("$obj")),
                code=code,
                ocode=cls.subcode("deserialize", opt, "$obj", ctx),
            )
        return Lambda(code)

    @ovld(priority=STD)
    def schema(self, t: type[UnionAlias], ctx: Context, /):
        options = get_args(t)
        return {"oneOf": [recurse(opt, ctx) for opt in options]}

    ##########################
    # Implementations: Enums #
    ##########################

    @code_generator_wrap_error(priority=STD3)
    def serialize(cls, t: type[Enum], obj: Enum, ctx: Context, /):
        return Lambda(Code("$obj.value"))

    @code_generator_wrap_error(priority=STD3)
    def deserialize(cls, t: type[Enum], obj: Any, ctx: Context, /):
        (t,) = get_args(t)
        return Lambda(Code("$t($obj)", t=t))

    @ovld(priority=STD3)
    def schema(self, t: type[Enum], ctx: Context, /):
        return {"enum": [e.value for e in t]}

    ##################################
    # Implementations: Literal Enums #
    ##################################

    @ovld(priority=STD)
    def serialize(self, t: type[IsLiteral], obj: Any, ctx: Context, /):
        options = get_args(t)
        if obj not in options:
            raise ValidationError(f"'{obj}' is not a valid option for {t}", ctx=ctx)
        return obj

    @ovld(priority=STD)
    def deserialize(self, t: type[IsLiteral], obj: Any, ctx: Context, /):
        options = get_args(t)
        if obj not in options:
            raise ValidationError(f"'{obj}' is not a valid option for {t}", ctx=ctx)
        return obj

    @ovld(priority=STD)
    def schema(self, t: type[IsLiteral], ctx: Context, /):
        return {"enum": list(get_args(t))}

    ##########################
    # Implementations: Dates #
    ##########################

    @code_generator_wrap_error(priority=STD)
    def deserialize(cls, t: type[datetime], obj: int | float, ctx: Context, /):
        return Lambda(Code("$fromtimestamp($obj)", fromtimestamp=datetime.fromtimestamp))

    # We specify schemas explicitly because they have special formats
    # The serialization/deserializable is taken care of by their model()

    @ovld(priority=STD)
    def schema(self, t: type[date], ctx: Context, /):
        return {"type": "string", "format": "date"}

    @ovld(priority=STD)
    def schema(self, t: type[datetime], ctx: Context, /):
        return {"type": "string", "format": "date-time"}

    #########################
    # Implementations: Path #
    #########################

    @ovld(priority=STD)
    def serialize(self, t: type[Path], obj: Path, ctx: Context):
        if isinstance(ctx, WorkingDirectory) and not obj.is_absolute():
            obj = obj.relative_to(ctx.directory)
        return str(obj)

    @ovld(priority=STD)
    def deserialize(self, t: type[Path], obj: str, ctx: Context):
        pth = Path(obj).expanduser()
        if isinstance(ctx, WorkingDirectory):
            pth = ctx.directory / pth
        return pth

    @ovld(priority=STD)
    def schema(self, t: type[Path], ctx: Context, /):
        return {"type": "string"}



from collections import defaultdict

import pytest


@pytest.mark.hookwrapper
def pytest_benchmark_group_stats(config, benchmarks, group_by):
    outcome = yield
    results = defaultdict(list)
    for bench in benchmarks:
        iface = bench["params"]["interface"].__name__
        group = bench["name"].replace(iface, "").replace(",]", "]").replace("[]", "")
        results[group].append(bench)
    outcome.force_result(results.items())



from dataclasses import dataclass
from numbers import Number
from typing import Literal

from serieux.model import field_at, model

from .common import has_312_features
from .definitions import Job, Pig, Point, Tree, Worker


def test_model_cached():
    ptm1 = model(Point)
    ptm2 = model(Point)
    assert ptm1 is ptm2


def test_model_recursive():
    tm = model(Tree)
    fleft = tm.fields[0]
    assert fleft.name == "left"
    assert fleft.type == Tree | Number


@has_312_features
def test_model_recursive_parametric():
    from .definitions_py312 import Tree

    tm = model(Tree[int])
    fleft = tm.fields[0]
    assert fleft.name == "left"
    assert fleft.type == Tree[int] | int


def test_model_default():
    assert model(int) is None


def test_model_none_default():
    m = model(Worker)
    assert len(m.fields) == 2
    assert m.fields[0].name == "name"
    assert m.fields[0].type is str
    assert m.fields[1].name == "job"
    assert m.fields[1].type == Job | None


@dataclass
class ColorBox:
    color: Literal["red", "green", "blue"]
    intensity: int = 100


def test_model_literal():
    m = model(ColorBox)
    assert len(m.fields) == 2
    assert m.fields[0].name == "color"
    assert m.fields[0].type == Literal["red", "green", "blue"]
    assert m.fields[1].name == "intensity"
    assert m.fields[1].type is int


def test_field_descriptions():
    m = model(Pig)
    p, w, b = m.fields

    assert p.name == "pinkness"
    assert p.description == "How pink the pig is"

    assert w.name == "weight"
    assert w.description == "Weight of the pig, in kilograms"

    assert b.name == "beautiful"
    assert b.description == "Is the pig...\ntruly...\n...beautiful?"


def test_eval_annotations_cross_file():
    from .lilmod.base import Point
    from .lilmod.ext import DrawingLine

    mdl = model(DrawingLine)
    assert {f.name: f.type for f in mdl.fields} == {"p1": Point, "p2": Point, "thickness": int}


def test_field_at():
    fld1 = field_at(dict[str, Worker], ["Jonathan", "job"])
    assert fld1.type == Job | None

    fld2 = field_at(dict[str, Worker], ["Jonathan", "job", "title"])
    assert fld2.name == "title"
    assert fld2.type is str

    fld3 = field_at(dict[str, Worker], ["Jonathan", "job", "gluglu"])
    assert fld3 is None

    fld4 = field_at(object, ["Jonathan", "job", "gluglu"])
    assert fld4 is None



import sys

from .ctx import Context


def _color(code, text):
    return f"\u001b[1m\u001b[{code}m{text}\u001b[0m"


def access_string(ctx):
    acc = getattr(ctx, "access_path", ["???"])
    return "".join([f".{field}" for field in acc]) if acc else "(at root)"


def context_string(
    ctx,
    message,
    show_source=True,
    source_context=1,
    indent=0,
    ellipsis_cutoff=3,
):
    n = 3 + indent
    acc_string = access_string(ctx)
    return_lines = [f"{_color(33, acc_string)}: {message}"]
    if show_source and (location := getattr(ctx, "location", None)):
        (l1, c1), (l2, c2) = location.linecols
        if c2 == 0:
            l2 -= 1
            c2 = 10_000_000_000_000
        lines = location.code.split("\n")
        start = l1 - source_context
        while start < 0 or not lines[start].strip():
            start += 1
        end = l2 + source_context
        while end >= len(lines) or not lines[end].strip():  # pragma: no cover
            end -= 1

        return_lines.append(f"{'':{indent}}@ {location.source}:{l1 + 1}")
        for li in range(start, end + 1):
            line = lines[li]
            if li == l2 and not line.strip():  # pragma: no cover
                break
            if li == l1 + ellipsis_cutoff and li < l2:
                return_lines.append(f"{'':{n}}  ...")
                continue
            elif li > l1 + ellipsis_cutoff and li < l2:
                continue

            hls = hle = 0
            if li == l1:
                hls = c1
            if li >= l1 and li < l2:
                hle = len(line)
            if li == l2:
                hle = c2

            if hls or hle:
                line = line[:hls] + _color(31, line[hls:hle]) + line[hle:]

            return_lines.append(f"{li + 1:{n}}: {line}")
    return "\n".join(return_lines)


def display_context(*args, file=sys.stdout, **kwargs):
    cs = context_string(*args, **kwargs)
    print(cs, file=file)


def merge_errors(*errors):
    collected = []
    for err in errors:
        if isinstance(err, ExceptionGroup):
            collected.extend(err.exceptions)
        elif isinstance(err, Exception):
            collected.append(err)
    return ValidationExceptionGroup("Some errors occurred", collected) if collected else None


class SerieuxError(Exception):
    pass


class IndividualSerieuxError(SerieuxError):
    def __init__(self, message=None, *, ctx=None):
        super().__init__(message)
        self.ctx = ctx
        if self.ctx is None:
            frame = sys._getframe(1)
            while frame:
                if "ctx" in frame.f_locals:
                    if isinstance(ctx_val := frame.f_locals["ctx"], Context):
                        self.ctx = ctx_val
                        break
                frame = frame.f_back

    @property
    def message(self):
        return self.args[0]

    def access_string(self):
        return access_string(self.ctx)

    def display(self, file=sys.stderr, prefix=""):
        print(prefix, end="", file=file)
        display_context(
            self.ctx,
            show_source=True,
            message=self.message,
            file=file,
            indent=2,
        )

    def __str__(self):
        location = getattr(self.ctx, "location", None)
        if location:
            (l1, c1), (l2, c2) = location.linecols
            lc = f"{l1}:{c1}-{l2}:{c2}" if l1 != l2 else f"{l1}:{c1}-{c2}"
            return f"{location.source}:{lc} -- {self.message}"
        else:
            return f"At path {access_string(self.ctx)}: {self.message}"


class NotGivenError(IndividualSerieuxError):
    pass


class ValidationExceptionGroup(SerieuxError, ExceptionGroup):
    def derive(self, excs):  # pragma: no cover
        return ValidationExceptionGroup(self.message, excs)

    def display(self, file=sys.stderr):
        for i, exc in enumerate(self.exceptions):
            exc.display(file=file, prefix=f"[#{i}] ")


class ValidationError(IndividualSerieuxError):
    def __init__(self, message=None, *, exc=None, ctx=None):
        if message is None:
            message = f"{type(exc).__name__}: {exc}"
        super().__init__(message=message, ctx=ctx)
        self.exc = exc


class SchemaError(IndividualSerieuxError):
    def __init__(self, message=None, *, exc=None, ctx=None):
        if message is None:  # pragma: no cover
            message = f"{type(exc).__name__}: {exc}"
        super().__init__(message=message, ctx=ctx)
        self.exc = exc



from dataclasses import dataclass
from types import UnionType
from typing import Annotated, Literal, TypeVar, Union, get_args, get_origin

T = TypeVar("T")


class InstructionMC(type):
    def __rmatmul__(cls, t):
        return Annotated[t, cls]

    def decompose(cls, t, all=False):
        method = cls.extract_all if all else cls.extract
        return strip(t, cls), method(t)

    def extract(cls, t):
        for x in cls.extract_all(t):
            return x
        return None

    def extract_all(cls, t):
        if get_origin(t) is not Annotated:
            return
        for x in t.__metadata__:
            if isinstance(x, cls):
                yield x

    @property
    def annotation_priority(cls):
        return 1


class BaseInstruction(metaclass=InstructionMC):
    def strip(self, t):
        return strip(t, self)

    def __rmatmul__(self, t):
        return Annotated[t, self]


@dataclass(frozen=True)
class Instruction(BaseInstruction):
    name: str
    annotation_priority: int = 1
    inherit: bool = True

    def __getitem__(self, t):
        return Annotated[t, self]

    def __str__(self):
        return self.name

    __repr__ = __str__


def annotate(cls, annotations):
    if get_origin(annotations) is Annotated:
        annotations = annotations.__metadata__
    if not annotations:
        return cls
    else:
        return Annotated[(cls, *annotations)]


def strip(cls, to_remove=None):
    def should_remove(a):
        if to_remove is None:
            return True
        if isinstance(to_remove, type):
            return isinstance(a, to_remove)
        else:
            return a == to_remove

    if get_origin(cls) is not Annotated:
        return cls
    anns = [a for a in cls.__metadata__ if not should_remove(a)]
    return annotate(get_args(cls)[0], anns)


def inherit(cls, target):
    if get_origin(cls) is not Annotated:
        return target
    _, *instrs = get_args(cls)
    new_instrs = [a for a in instrs if getattr(a, "inherit", False)]
    if not new_instrs:
        return target
    return Annotated[(target, *new_instrs)]


def pushdown(cls):
    if get_origin(cls) is not Annotated:
        return cls
    typ, *instrs = get_args(cls)
    new_instrs = [a for a in instrs if getattr(a, "inherit", False)]
    if not new_instrs:
        return typ
    if (orig := get_origin(typ)) and orig is not Literal:
        args = get_args(typ)
        if orig is UnionType:
            orig = Union
        return orig[tuple([Annotated[(a, *new_instrs)] for a in args])]
    else:
        return typ



from __future__ import annotations

from dataclasses import dataclass, field
from enum import Enum
from numbers import Number


@dataclass
class Tree:
    left: Tree | Number
    right: Tree | Number


@dataclass
class Citizen:
    name: str
    birthyear: int
    hometown: str


@dataclass
class Country:
    languages: list[str]
    capital: str
    population: int
    citizens: list[Citizen]


@dataclass
class World:
    countries: dict[str, Country]


@dataclass
class Point:
    x: int
    y: int


@dataclass
class Point3D(Point):
    z: int


class Color(str, Enum):
    RED = "red"
    GREEN = "green"
    BLUE = "blue"


class Level(Enum):
    HI = 2
    MED = 1
    LO = 0


@dataclass
class Pig:
    # How pink the pig is
    pinkness: float

    weight: float
    """Weight of the pig, in kilograms"""

    # Is the pig...
    # truly...
    beautiful: bool = True  # ...beautiful?


@dataclass
class Defaults:
    name: str
    aliases: list[str] = field(default_factory=list)
    cool: bool = field(default=False, kw_only=True)


@dataclass
class Player:
    first: str
    last: str
    batting: float


@dataclass
class Team:
    name: str
    players: list[Player]


@dataclass
class Job:
    # Name of the job
    title: str
    # How much it pays, in dollars
    yearly_pay: float


@dataclass
class Worker:
    name: str
    job: Job = None


DID = dict[str, "int | DID"]


@dataclass
class DIDHolder:
    did: DID


ListTree = list["Point | ListTree"]


@dataclass
class LTHolder:
    lt: ListTree


class DotDict(dict):
    def __getattr__(self, attr):
        return self[attr]



import ast
import inspect
import re
import tokenize
from dataclasses import dataclass
from functools import partial
from textwrap import dedent

#############################
# Extracting the docstrings #
#############################


def scrape_comments(src: str):
    """Use the tokenizer module to fetch comments from source.

    Arguments:
        src: The source code.

    Returns:
        List of (line, column, "COMMENT", comment_string)
    """
    lines = bytes(src, encoding="utf8").splitlines(keepends=True)
    return [
        (*tok.start, "COMMENT", tok.string[1:].strip())
        for tok in tokenize.tokenize(partial(next, iter(lines)))
        if tok.type == tokenize.COMMENT
    ]


class AttributeVisitor(ast.NodeVisitor):
    """Walk an AST to gather variable and docstring tokens and their lines/columns."""

    def __init__(self):
        self.data = []
        self.prefix = None

    def add_data(self, node, kind, content):
        self.data.append((node.lineno, node.col_offset, kind, content))

    def visit_body(self, name, stmts):
        old_prefix = self.prefix
        if self.prefix is None:
            self.prefix = ""
        else:
            self.prefix += f"{name}."
        for stmt in stmts:
            if (
                isinstance(stmt, ast.Expr)
                and isinstance(stmt.value, ast.Constant)
                and isinstance(stmt.value.value, str)
            ):
                self.add_data(stmt, "DOC", stmt.value.value)
            else:
                self.visit(stmt)
        self.prefix = old_prefix

    def visit_ClassDef(self, node):
        if self.prefix is not None:  # pragma: no cover
            self.add_data(node, "VARIABLE", f"{self.prefix}{node.name}")
        self.visit_body(node.name, node.body)

    def visit_FunctionDef(self, node):
        if self.prefix is not None:
            self.add_data(node, "VARIABLE", f"{self.prefix}{node.name}")
        for arg in [*node.args.posonlyargs, *node.args.args, *node.args.kwonlyargs]:
            self.add_data(arg, "ARGUMENT", f"{self.prefix}{arg.arg}" if self.prefix else arg.arg)
        self.visit_body(node.name, node.body)

    def visit_Assign(self, node):  # pragma: no cover
        self.generic_visit(node, may_assign=True)

    def visit_AnnAssign(self, node):
        self.generic_visit(node, may_assign=True)

    def visit_Name(self, node):
        if isinstance(node.ctx, ast.Store):
            self.add_data(node, "VARIABLE", f"{self.prefix}{node.id}")

    def generic_visit(self, node, may_assign=False):
        if isinstance(node, ast.stmt) and not may_assign:
            self.add_data(node, "OTHER", None)
        super().generic_visit(node)


def scrape_variables_and_docstrings(src: str):
    """Scrape the variables and docstrings from source code.

    Arguments:
        src: The source code.

    Returns:
        List of (line, column, "VARIABLE"|"DOC"|"OTHER", name_or_content)
    """
    visitor = AttributeVisitor()
    visitor.visit(ast.parse(src))
    return visitor.data


_cached_docstrings = {}


@dataclass
class VariableDoc:
    doc: str
    metadata: dict[str, object]

    @staticmethod
    def from_lines(lines):
        metadata = {}
        doclines = []
        for line in lines:
            sl = line.strip()
            if sl.startswith("[") and sl.endswith("]"):
                for match in re.finditer(r"\[([^:]+)(?::\s*([^\[]+))?\]", line):
                    key, value = match.groups()
                    if value is not None:
                        try:
                            value = float(value)
                        except ValueError:
                            pass
                    metadata[key] = value if value is not None else True
            else:
                doclines.append(line)

        return VariableDoc(
            doc="\n".join(doclines),
            metadata=metadata,
        )


def get_variable_data(cls):
    """Get the docstrings for individual attributes of a class.

    Arguments:
        cls: The class for which we want to get attribute documentation.

    Returns:
        A dict from variable name to its associated docstring (after itself) and/or
        comment (above itself).
    """
    if cls in _cached_docstrings:
        return _cached_docstrings[cls]
    if hasattr(cls, "__variable_data__"):
        return cls.__variable_data__
    docs = {}
    current = None
    current_line = None
    for_next = []
    try:
        src = dedent(inspect.getsource(cls))
    except (OSError, TypeError):
        return {}
    # We concatenate comment tokens from the tokenizer with
    # variable/docstring tokens extracted using the ast module
    data = sorted(scrape_comments(src) + scrape_variables_and_docstrings(src))
    current_kind = None
    for line, _, kind, content in data:
        if kind == "COMMENT":
            if current is not None and current_line == line:
                docs[current].append(content)
            else:
                for_next.append(content)
        elif kind == "DOC" and current_kind == "VARIABLE":
            docs[current].append(content)
        elif kind == "VARIABLE" or kind == "ARGUMENT":
            if content not in docs:
                docs[content] = for_next
            for_next = []
            current = content
            current_kind = kind
            current_line = line
        elif kind == "OTHER":
            current = current_line = None
            current_kind = None
            for_next = []
    rval = _cached_docstrings[cls] = {
        k: VariableDoc.from_lines(lines) for k, lines in docs.items()
    }
    return rval


def get_attribute_docstrings(cls):
    results = {}
    for subcls in cls.mro():
        results.update(get_variable_data(subcls))
    return results



from typing import TYPE_CHECKING, Annotated, Any, TypeAlias

from ovld import Medley, call_next, ovld, recurse

from ..ctx import Context
from ..instructions import Instruction, T
from ..priority import HI5
from .proxy import LazyProxy

#############
# Constants #
#############


if TYPE_CHECKING:
    Lazy: TypeAlias = Annotated[T, None]
    DeepLazy: TypeAlias = Annotated[T, None]
else:
    Lazy = Instruction("Lazy", annotation_priority=2, inherit=False)
    DeepLazy = Instruction("DeepLazy", annotation_priority=2, inherit=True)


###################
# Implementations #
###################


class LazyDeserialization(Medley):
    @ovld(priority=HI5)
    def deserialize(self, t: type[Any @ Lazy], value: object, ctx: Context):
        def evaluate():
            return recurse(Lazy.strip(t), value, ctx)

        return LazyProxy(evaluate, type=t)

    @ovld(priority=HI5)
    def deserialize(self, t: type[Any @ DeepLazy], value: object, ctx: Context):
        def evaluate():
            return call_next(t, value, ctx)

        return LazyProxy(evaluate, type=t)

    @ovld  # pragma: no cover
    def deserialize(self, t: Any, value: LazyProxy, ctx: Context):
        return recurse(t, value._obj, ctx)



from dataclasses import dataclass
from typing import Annotated

import pytest

from serieux import Serieux
from serieux.exc import ValidationError
from serieux.features.comment import Comment, CommentedObjects, CommentRec, comment_field
from serieux.features.proxy import CommentProxy
from serieux.features.tagset import value_field

featured = (Serieux + CommentedObjects)()
serialize = featured.serialize
deserialize = featured.deserialize


@dataclass
class Person:
    name: str
    age: int


def test_comment_serialize_int():
    proxy = CommentProxy(42, "The answer")
    serialized = serialize(Comment[int, str], proxy)
    expected = {value_field: 42, comment_field: "The answer"}
    assert serialized == expected


def test_comment_deserialize_int():
    serialized = {value_field: 42, comment_field: "The answer"}
    deserialized = deserialize(Comment[int, str], serialized)
    assert isinstance(deserialized, CommentProxy)
    assert deserialized._obj == 42
    assert deserialized._ == "The answer"


def test_comment_serialize_person():
    person = Person(name="Alice", age=30)
    proxy = CommentProxy(person, "A nice person")
    serialized = serialize(Comment[Person, str], proxy)
    expected = {"name": "Alice", "age": 30, comment_field: "A nice person"}
    assert serialized == expected


def test_comment_deserialize_person():
    serialized = {"name": "Alice", "age": 30, comment_field: "A nice person"}
    deserialized = deserialize(Comment[Person, str], serialized)
    assert isinstance(deserialized, CommentProxy)
    assert isinstance(deserialized._obj, Person)
    assert deserialized._obj == Person(name="Alice", age=30)
    assert deserialized._ == "A nice person"


def test_comment_with_complex_comment_type():
    comment_person = Person("Bob", 25)
    proxy = CommentProxy(100, comment_person)

    serialized = serialize(Comment[int, Person], proxy)
    expected = {value_field: 100, comment_field: {"name": "Bob", "age": 25}}
    assert serialized == expected

    deserialized = deserialize(Comment[int, Person], serialized)
    assert isinstance(deserialized, CommentProxy)
    assert deserialized._obj == 100
    assert deserialized._ == comment_person


def test_comment_without_comment_field():
    data = {"name": "Charlie", "age": 35}
    result = deserialize(Comment[Person, str], data)
    assert type(result) is Person
    assert result.name == "Charlie"
    assert result.age == 35


def test_comment_required_not_commentproxy():
    person = Person("Eve", 28)
    with pytest.raises(ValidationError):
        serialize(Annotated[Person, Comment(str, True)], person)


def test_comment_required_without_comment():
    data = {"name": "Charlie", "age": 35}
    with pytest.raises(ValidationError):
        deserialize(Annotated[Person, Comment(str, True)], data)


def test_comment_required_without_comment_int():
    data = 100
    with pytest.raises(ValidationError):
        deserialize(Annotated[int, Comment(str, True)], data)


def test_comment_serialize_non_proxy():
    person = Person("David", 40)
    serialized = serialize(Comment[Person, str], person)
    expected = {"name": "David", "age": 40}
    assert serialized == expected


def test_comment_serialize_proxy_to_normal_type():
    person = Person("David", CommentProxy(40, "v. young"))
    serialized = serialize(Person, person)
    expected = {"name": "David", "age": 40}
    assert serialized == expected


def test_comment_schema(schematest):
    schematest(
        type=Comment[Person, str],
        value=CommentProxy(Person("Alice", 30), "note"),
    )


def test_comment_schema_required(schematest):
    schematest(
        type=Annotated[Person, Comment(str, required=True)],
        value=CommentProxy(Person("Bob", 25), "required comment"),
    )


def test_comment_schema_primitive(schematest):
    schematest(
        type=Annotated[str, Comment(str, required=True)],
        value=CommentProxy("hello", "world"),
    )


def test_commentrec_serialize():
    person = Person("David", CommentProxy(40, "v. young"))

    serialized = serialize(CommentRec[Person, str], person)
    expected = {"name": "David", "age": {value_field: 40, comment_field: "v. young"}}
    assert serialized == expected

    deserialized = deserialize(CommentRec[Person, str], expected)
    assert deserialized == person
    assert deserialized.age._ == "v. young"



import base64
import json
from dataclasses import dataclass
from hashlib import sha256
from typing import Any

from cryptography.fernet import Fernet
from ovld import Medley, call_next
from ovld.dependent import Regexp
from rich.pretty import pprint

from serieux import Context, Serieux, deserialize, serialize
from serieux.exc import ValidationError
from serieux.instructions import Instruction

##################
# Implementation #
##################


Secret = Instruction("Secret")


class EncryptionKey(Context):
    password: str
    key: Fernet = None

    def __post_init__(self):
        encoded = base64.b64encode(sha256(self.password.encode()).digest())
        self.key = Fernet(encoded)

    def encrypt(self, value):
        encrypted = self.key.encrypt(json.dumps(value).encode("utf8")).decode("utf8")
        return f"~CRYPT~{encrypted}"

    def decrypt(self, encrypted: str):
        return json.loads(self.key.decrypt(encrypted.lstrip("~CRYPT~")))


@Serieux.extend
class Encrypt(Medley):
    def serialize(self, t: type[Any @ Secret], obj: object, ctx: EncryptionKey):
        result = call_next(Secret.strip(t), obj, ctx - EncryptionKey)
        return ctx.encrypt(result)

    def deserialize(self, t: type[Any @ Secret], obj: Regexp[r"^~CRYPT~.*"], ctx: EncryptionKey):
        obj = ctx.decrypt(obj)
        return call_next(Secret.strip(t), obj, ctx - EncryptionKey)


#################
# Demonstration #
#################


@dataclass
class User:
    name: str
    passwords: Secret[dict[str, str]]


autoinput = {
    "Password: ": "bonjour",
    "Enter password again: ": "bonjour",
}.__getitem__


def main(input=autoinput):
    olivier = User(name="olivier", passwords={"google": "tobeornottobeevil", "apple": "banana"})

    print("\n== Original ==\n")
    pprint(olivier)

    password = input("Password: ")
    ctx = EncryptionKey(password)

    serial = serialize(User, olivier, ctx)

    print("\n== Serialized ==\n")
    pprint(serial)

    # Our secrets are safe!
    assert "tobeornottobeevil" not in json.dumps(serial)
    assert "banana" not in json.dumps(serial)

    password = input("Enter password again: ")
    ctx = EncryptionKey(password)

    try:
        olivier2 = deserialize(User, serial, ctx)
    except ValidationError:
        print("Invalid password!")
        return

    print("\n== Deserialized ==\n")
    pprint(olivier2)

    assert olivier == olivier2


# This is for the regression tests we run on the examples with pytest
# Fernet uses a random component, so the output is not deterministic
main.do_not_test_output = True


if __name__ == "__main__":
    main(input)



from dataclasses import dataclass, field
from itertools import count
from random import (
    choice,
    expovariate as exp,  # noqa: F401
    normalvariate as normal,  # noqa: F401
    randint,  # noqa: F401
    random as rand,  # noqa: F401
    seed,
    uniform,  # noqa: F401
)

from ovld import Medley, ovld, recurse
from ovld.dependent import Regexp
from rich.pretty import pprint

from serieux import Serieux, deserialize
from serieux.ctx import AccessPath, Context

##################
# Implementation #
##################


@dataclass
class Tracker:
    length: int
    current: int = 0

    def advance(self):
        self.current = (self.current + 1) % self.length
        return self.current == 0


class Grid(AccessPath):
    trackers: dict[str, Tracker] = field(default_factory=dict)

    def get_choice(self, choices):
        if (acc := self.access_path) not in self.trackers:
            self.trackers[acc] = Tracker(len(choices))
        tracker = self.trackers[acc]
        assert tracker.length == len(choices)
        return choices[tracker.current]

    def advance(self):
        for tracker in self.trackers.values():
            if not tracker.advance():
                return True
        else:
            return False


@Serieux.extend
class Sampler(Medley):
    def deserialize(self, typ: type[object], obj: Regexp["^~"], ctx: Context):
        return eval(obj.lstrip("~"), globals())

    @ovld(priority=-1)
    def deserialize(self, typ: type[object], obj: list, ctx: Context):
        return recurse(typ, choice(obj), ctx)

    @ovld(priority=-1)
    def deserialize(self, typ: type[object], obj: list, ctx: Grid):
        return recurse(typ, ctx.get_choice(obj), ctx)


#################
# Demonstration #
#################


@dataclass
class Config:
    lr: float
    model: str
    dataset: str


def main():
    seed(1234)
    cfg = {
        "lr": "~exp(1)",
        "model": ["ConvNet", "AutoEncoder", "LLM"],
        "dataset": ["MNIST", "ImageNet", "CIFAR-10"],
    }
    g = Grid()
    for i in count():
        print(f"\n== Iteration {i} ==")
        pprint(g.trackers)
        pprint(deserialize(Config, cfg, g))
        if not g.advance():
            break


if __name__ == "__main__":
    main()



from dataclasses import dataclass
from typing import Annotated, Any

from ovld import Medley, call_next, ovld, recurse

from ..ctx import Context
from ..exc import ValidationError
from ..instructions import BaseInstruction
from ..priority import HI6, STD
from ..schema import AnnotatedSchema
from .proxy import CommentProxy
from .tagset import value_field

#############
# Constants #
#############


comment_field = "$comment"


@dataclass(frozen=True)
class Comment(BaseInstruction):
    comment_type: type
    required: bool = False
    inherit: bool = False

    def __class_getitem__(cls, args):
        t, ct = args
        return Annotated[t, cls(ct)]


@dataclass(frozen=True)
class CommentRec(Comment):
    inherit: bool = True


###################
# Implementations #
###################


class CommentedObjects(Medley):
    @ovld(priority=HI6)
    def serialize(self, t: type[Any @ Comment], obj: CommentProxy, ctx: Context):
        base, instr = Comment.decompose(t)
        rval = recurse(base, obj._obj, ctx)
        comment = recurse(instr.comment_type, obj._, ctx)
        if not isinstance(rval, dict):
            rval = {value_field: rval}
        rval[comment_field] = comment
        return rval

    @ovld(priority=HI6)
    def serialize(self, t: type[Any @ Comment], obj: object, ctx: Context):
        instr = Comment.extract(t)
        if instr.required:
            raise ValidationError("Comment is required but object is not a CommentProxy", ctx=ctx)
        return call_next(t, obj, ctx)

    @ovld(priority=STD)
    def serialize(self, t: Any, obj: CommentProxy, ctx: Context):
        return recurse(t, obj._obj, ctx)

    @ovld(priority=HI6)
    def deserialize(self, t: type[Any @ Comment], obj: dict, ctx: Context):
        base, instr = Comment.decompose(t)
        if comment_field not in obj:
            if instr.required:
                raise ValidationError(
                    f"Comment is required but '{comment_field}' field is missing", ctx=ctx
                )
            return call_next(t, obj, ctx)
        obj = dict(obj)
        comment = obj.pop(comment_field)
        if value_field in obj:
            obj = obj.pop(value_field)
        main = recurse(base, obj, ctx)
        comment = recurse(instr.comment_type, comment, ctx)
        return CommentProxy(main, comment)

    @ovld(priority=HI6)
    def deserialize(self, t: type[Any @ Comment], obj: object, ctx: Context):
        instr = Comment.extract(t)
        if instr.required:
            raise ValidationError(
                f"Comment is required but input is not a dictionary with '{comment_field}' field",
                ctx=ctx,
            )
        return call_next(t, obj, ctx)

    @ovld
    def schema(self, t: type[Any @ Comment], ctx: Context):
        base, instr = Comment.decompose(t)
        base_schema = recurse(base, ctx)
        comment_schema = recurse(instr.comment_type, ctx)
        if base_schema.get("type", None) == "object":
            return AnnotatedSchema(
                base_schema,
                properties={comment_field: comment_schema},
                required=[comment_field] if instr.required else [],
            )
        else:
            required = [value_field]
            if instr.required:
                required.append(comment_field)
            return {
                "type": "object",
                "properties": {
                    value_field: base_schema,
                    comment_field: comment_schema,
                },
                "required": required,
            }



from dataclasses import dataclass
from types import NoneType
from typing import Literal

import pytest

from serieux import schema as _schema
from serieux.model import Extensible
from serieux.schema import AnnotatedSchema, Schema

from .common import has_312_features
from .definitions import Color, Defaults, LTHolder, Pig, Point


def schema(t, root=False, ref_policy="norepeat"):
    return _schema(t).compile(root=root, ref_policy=ref_policy)


def test_schema_hashable():
    sch1 = Schema(int)
    sch2 = Schema(int)
    assert sch1 == sch1
    assert not (sch1 == sch2)
    assert {sch1: 1, sch2: 2} == {sch1: 1, sch2: 2}


def test_schema_int():
    assert schema(int) == {"type": "integer"}


def test_schema_bool():
    assert schema(bool) == {"type": "boolean"}


def test_schema_str():
    assert schema(str) == {"type": "string"}


def test_schema_None():
    assert schema(NoneType) == {"type": "null"}


def test_schema_enum():
    assert schema(Color) == {"enum": ["red", "green", "blue"]}


def test_schema_literal():
    assert schema(Literal["red", "green", "blue"]) == {"enum": ["red", "green", "blue"]}


def test_schema_list():
    assert schema(list[int]) == {"type": "array", "items": {"type": "integer"}}


def test_schema_dict():
    assert schema(dict[str, float]) == {
        "type": "object",
        "additionalProperties": {"type": "number"},
    }


def test_schema_dict_non_str_keys():
    with pytest.raises(Exception, match="Cannot create a schema for dicts with non-string keys"):
        schema(dict[int, str])


def test_schema_nested():
    assert schema(dict[str, list[int]]) == {
        "type": "object",
        "additionalProperties": {"type": "array", "items": {"type": "integer"}},
    }


def test_schema_dataclass():
    assert schema(Point) == {
        "type": "object",
        "properties": {"x": {"type": "integer"}, "y": {"type": "integer"}},
        "required": ["x", "y"],
        "additionalProperties": False,
    }


def test_schema_extensible_dataclass():
    assert schema(Extensible[Point]) == {
        "type": "object",
        "properties": {"x": {"type": "integer"}, "y": {"type": "integer"}},
        "required": ["x", "y"],
        "additionalProperties": True,
    }


def test_schema_dataclass_2():
    assert schema(Defaults) == {
        "type": "object",
        "properties": {
            "name": {"type": "string"},
            "aliases": {"type": "array", "items": {"type": "string"}},
            "cool": {"type": "boolean"},
        },
        "required": ["name"],
        "additionalProperties": False,
    }


@has_312_features
def test_schema_recursive():
    from .definitions_py312 import Tree

    assert schema(Tree[int]) == {
        "type": "object",
        "properties": {
            "left": {
                "oneOf": [
                    {"$ref": "#"},
                    {"type": "integer"},
                ]
            },
            "right": {
                "oneOf": [
                    {"$ref": "#"},
                    {"type": "integer"},
                ]
            },
        },
        "required": ["left", "right"],
        "additionalProperties": False,
    }


@has_312_features
def test_schema_recursive_policy_always():
    from .definitions_py312 import Tree

    assert schema(Tree[int], root=True, ref_policy="always") == {
        "$schema": "https://json-schema.org/draft/2020-12/schema",
        "$ref": "#/$defs/Tree",
        "$defs": {
            "Tree": {
                "type": "object",
                "properties": {
                    "left": {
                        "oneOf": [
                            {"$ref": "#/$defs/Tree"},
                            {"type": "integer"},
                        ]
                    },
                    "right": {
                        "oneOf": [
                            {"$ref": "#/$defs/Tree"},
                            {"type": "integer"},
                        ]
                    },
                },
                "required": ["left", "right"],
                "additionalProperties": False,
            }
        },
    }


@has_312_features
def test_schema_recursive_policy_two_trees():
    from .definitions_py312 import Tree

    @dataclass
    class DoubleTree:
        it: Tree[int]
        st: Tree[str]

    assert schema(DoubleTree, root=True, ref_policy="always") == {
        "$schema": "https://json-schema.org/draft/2020-12/schema",
        "$ref": "#/$defs/DoubleTree",
        "$defs": {
            "DoubleTree": {
                "type": "object",
                "properties": {
                    "it": {"$ref": "#/$defs/Tree"},
                    "st": {"$ref": "#/$defs/Tree2"},
                },
                "required": ["it", "st"],
                "additionalProperties": False,
            },
            "Tree": {
                "type": "object",
                "properties": {
                    "left": {
                        "oneOf": [
                            {"$ref": "#/$defs/Tree"},
                            {"type": "integer"},
                        ]
                    },
                    "right": {
                        "oneOf": [
                            {"$ref": "#/$defs/Tree"},
                            {"type": "integer"},
                        ]
                    },
                },
                "required": ["left", "right"],
                "additionalProperties": False,
            },
            "Tree2": {
                "type": "object",
                "properties": {
                    "left": {
                        "oneOf": [
                            {"$ref": "#/$defs/Tree2"},
                            {"type": "string"},
                        ]
                    },
                    "right": {
                        "oneOf": [
                            {"$ref": "#/$defs/Tree2"},
                            {"type": "string"},
                        ]
                    },
                },
                "required": ["left", "right"],
                "additionalProperties": False,
            },
        },
    }


@has_312_features
def test_schema_recursive_policy_never():
    from .definitions_py312 import Tree

    with pytest.raises(Exception, match="Recursive schema"):
        schema(Tree[int], root=True, ref_policy="never")


@dataclass
class TwoPoints:
    # First point
    a: Point
    # Second point
    b: Point


def test_schema_policy_never_minimal():
    never = schema(TwoPoints, ref_policy="never")
    minimal = schema(TwoPoints, ref_policy="minimal")
    assert (
        never
        == minimal
        == {
            "type": "object",
            "properties": {
                "a": {
                    "type": "object",
                    "description": "First point",
                    "properties": {"x": {"type": "integer"}, "y": {"type": "integer"}},
                    "required": ["x", "y"],
                    "additionalProperties": False,
                },
                "b": {
                    "type": "object",
                    "description": "Second point",
                    "properties": {"x": {"type": "integer"}, "y": {"type": "integer"}},
                    "required": ["x", "y"],
                    "additionalProperties": False,
                },
            },
            "required": ["a", "b"],
            "additionalProperties": False,
        }
    )


def test_schema_policy_norepeat():
    assert schema(TwoPoints, ref_policy="norepeat") == {
        "type": "object",
        "properties": {
            "a": {
                "type": "object",
                "description": "First point",
                "properties": {"x": {"type": "integer"}, "y": {"type": "integer"}},
                "required": ["x", "y"],
                "additionalProperties": False,
            },
            "b": {
                "$ref": "#/properties/a",
                "description": "Second point",
            },
        },
        "required": ["a", "b"],
        "additionalProperties": False,
    }


def test_schema_policy_always():
    assert schema(TwoPoints, ref_policy="always") == {
        "$ref": "#/$defs/TwoPoints",
        "$defs": {
            "Point": {
                "type": "object",
                "properties": {"x": {"type": "integer"}, "y": {"type": "integer"}},
                "required": ["x", "y"],
                "additionalProperties": False,
            },
            "TwoPoints": {
                "type": "object",
                "properties": {
                    "a": {"$ref": "#/$defs/Point", "description": "First point"},
                    "b": {"$ref": "#/$defs/Point", "description": "Second point"},
                },
                "required": ["a", "b"],
                "additionalProperties": False,
            },
        },
    }


def test_schema_descriptions():
    assert schema(Pig) == {
        "type": "object",
        "properties": {
            "pinkness": {"type": "number", "description": "How pink the pig is"},
            "weight": {"type": "number", "description": "Weight of the pig, in kilograms"},
            "beautiful": {
                "type": "boolean",
                "description": "Is the pig...\ntruly...\n...beautiful?",
                "default": True,
            },
        },
        "required": ["pinkness", "weight"],
        "additionalProperties": False,
    }


@pytest.mark.xfail
def test_schema_recursive_ltholder():
    assert schema(LTHolder) == {
        "type": "object",
        "properties": {
            "lt": {
                "type": "array",
                "items": {
                    "oneOf": [
                        {
                            "type": "object",
                            "properties": {"x": {"type": "integer"}, "y": {"type": "integer"}},
                            "required": ["x", "y"],
                            "additionalProperties": False,
                        },
                        {
                            "$ref": "#/properties/lt",
                        },
                    ]
                },
            },
        },
        "required": ["lt"],
        "additionalProperties": False,
    }


class AdditivePoint(Point):
    @classmethod
    def serieux_schema(cls, ctx, call_next):
        return AnnotatedSchema(
            parent=call_next(Point, ctx),
            properties={"more": {"type": "string"}},
            required=["more"],
        )


def test_schema_additive_point():
    assert schema(AdditivePoint) == {
        "type": "object",
        "properties": {
            "x": {"type": "integer"},
            "y": {"type": "integer"},
            "more": {"type": "string"},
        },
        "required": ["x", "y", "more"],
        "additionalProperties": False,
    }



import json
import os
import tomllib
from dataclasses import dataclass
from datetime import date
from pathlib import Path
from types import NoneType

import pytest

from serieux import Serieux
from serieux.ctx import AccessPath
from serieux.exc import ValidationError
from serieux.features.fromfile import IncludeFile, WorkingDirectory
from serieux.features.partial import Sources

from ..definitions import Citizen, Country, Player, Team, World

deserialize = (Serieux + IncludeFile)().deserialize

here = Path(__file__).parent
datapath = Path(__file__).parent.parent / "data"


def test_deserialize_from_file():
    assert deserialize(Country, datapath / "canada.yaml") == Country(
        languages=["English", "French"],
        capital="Ottawa",
        population=39_000_000,
        citizens=[
            Citizen(
                name="Olivier",
                birthyear=1985,
                hometown="Montreal",
            ),
            Citizen(
                name="Abraham",
                birthyear=2018,
                hometown="Shawinigan",
            ),
        ],
    )


def test_deserialize_override():
    srcs = Sources(
        datapath / "canada.yaml",
        {"capital": "Montreal"},
    )
    assert deserialize(Country, srcs) == Country(
        languages=["English", "French"],
        capital="Montreal",
        population=39_000_000,
        citizens=[
            Citizen(
                name="Olivier",
                birthyear=1985,
                hometown="Montreal",
            ),
            Citizen(
                name="Abraham",
                birthyear=2018,
                hometown="Shawinigan",
            ),
        ],
    )


def test_deserialize_world():
    world = deserialize(World, datapath / "world.yaml")
    assert world == World(
        countries={
            "canada": Country(
                languages=["English", "French"],
                capital="Ottawa",
                population=39_000_000,
                citizens=[
                    Citizen(
                        name="Olivier",
                        birthyear=1985,
                        hometown="Montreal",
                    ),
                    Citizen(
                        name="Abraham",
                        birthyear=2018,
                        hometown="Shawinigan",
                    ),
                ],
            ),
            "france": Country(
                languages=["French"],
                capital="Paris",
                population=68_000_000,
                citizens=[
                    Citizen(
                        name="Jeannot",
                        birthyear=1893,
                        hometown="Lyon",
                    ),
                ],
            ),
        }
    )


def test_deserialize_json():
    file = datapath / "world.json"
    # Sanity check that this is a valid JSON file
    json.loads(file.read_text())
    world = deserialize(World, file)
    world_baseline = deserialize(World, file.with_suffix(".yaml"))
    assert world == world_baseline


def test_deserialize_toml():
    file = datapath / "world.toml"
    # Sanity check that this is a valid TOML file
    tomllib.loads(file.read_text())
    world = deserialize(World, file)
    world_baseline = deserialize(World, file.with_suffix(".yaml"))
    assert world == world_baseline


def test_deserialize_missing_file():
    with pytest.raises(ValidationError, match="Could not read data"):
        deserialize(World, datapath / "missing.yaml")


def test_deserialize_read_direct():
    team = deserialize(Team, datapath / "team.yaml")
    assert team.players[0] == Player(first="Olivier", last="Breuleux", batting=0.9)


def test_deserialize_incomplete(check_error_display):
    with check_error_display("KeyError: 'capital'"):
        deserialize(Country, datapath / "france.yaml", AccessPath())


def test_deserialize_invalid(check_error_display):
    with check_error_display("Cannot deserialize string"):
        deserialize(Country, datapath / "invalid.yaml", AccessPath())


def test_deserialize_oops_world(check_error_display):
    with check_error_display("Cannot deserialize string"):
        deserialize(World, datapath / "oops-world.yaml", AccessPath())


def test_make_path_for(tmp_path):
    wd = WorkingDirectory(directory=tmp_path)
    path = wd.make_path_for(name="test_file", suffix=".txt")
    assert path.name == "test_file.txt"
    assert path.parent == tmp_path
    assert not path.exists()


def test_save(tmp_path):
    wd = WorkingDirectory(directory=tmp_path)
    txt = "Some delicious text"
    relpath = wd.save_to_file(txt, suffix=".txt")
    path = wd.directory / relpath
    assert path.exists()
    assert path.read_text() == txt


def test_save_bytes(tmp_path):
    wd = WorkingDirectory(directory=tmp_path)
    rbytes = os.urandom(100)
    relpath = wd.save_to_file(rbytes, suffix=".txt")
    path = wd.directory / relpath
    assert path.exists()
    assert path.read_bytes() == rbytes


def test_save_callback(tmp_path):
    li = []
    wd = WorkingDirectory(directory=tmp_path)
    relpath = wd.save_to_file(callback=li.append, suffix=".txt")
    path = wd.directory / relpath
    assert li == [path]


def test_wd_origin(tmp_path):
    origin = tmp_path / "xxx.yaml"
    wd = WorkingDirectory(origin=origin)
    assert wd.origin == origin
    assert wd.directory == tmp_path


@dataclass
class Datatypes:
    strong: str
    integger: int
    flowhat: float
    boule: bool
    nuttin: NoneType
    date: date | None


def test_deserialize_types():
    data = deserialize(Datatypes, datapath / "all.yaml")
    assert data == Datatypes(
        strong="hello", integger=5, flowhat=4.4, boule=True, nuttin=None, date=date(2025, 1, 3)
    )
